# -*- coding: utf-8 -*-
"""
Provides the config and the handling of it
"""
import copy
import os
import sys
import json


def get_config_dir():
    """
    Returns the configuration dictionary of the user

    Returns
    -------
    return : string
             config folder
             LINUX / OSX: ~/.config/swat-em/
             WINSOWS: ../appdata/swat-em/
    """
    if "linux" in sys.platform:
        user_config_dir = os.path.join(os.path.expanduser("~"), ".config/swat-em")
    elif "darwin" in sys.platform:
        user_config_dir = os.path.join(os.path.expanduser("~"), ".config/swat-em")
    else:
        user_config_dir = os.path.join(os.getenv("APPDATA"), "swat-em")
    return user_config_dir


def get_init_config():
    config = {}
    config["N_nu_el"] = 19
    config["N_nu_mech"] = 19
    config["threshold_MMF_harmonics"] = 0.01
    config["kw_min"] = 0.01
    config["num_MMF_points"] = 3601
    config["plot_MMF_harmonics"] = 0.15
    config["plt"] = {
        "lw": 2.0,
        "lw_thin": 1.0,
        "DPI": 90,
        "res": [800, 600],
        "phase_colors": [
            "#1b9e77",
            "#d95f02",
            "#7570b3",
            "#e7298a",
            "#66a61e",
            "#e6ab02",
            "#a6761d",
        ],  # http://colorbrewer2.org/#type=qualitative&scheme=Dark2&n=7
        "line_colors": [
            "#377eb8",
            "#e41a1c",
            "#4daf4a",
            "#984ea3",
            "#ff7f00",
            "#a65628",
            "#f781bf",
        ],  # http://colorbrewer2.org/#type=qualitative&scheme=Set1&n=8
        "magnet_colors": ["#FFA7A7", "#9EB0FF"],
        "magnet_linewidth": 2,
        "magnet_alpha_m": 0.85,
        "draw_poles": True,
    }
    config["report"] = {}

    config["radial_force"] = {"num_modes": 3}
    config["report_txt"] = {"font": "Monospace", "fontsize": 10}
    config["view"] = {
        "plot_tabs": [
            "tab_slot",
            "tab_layout_polar",
            "tab_overhang",
            "tab_star",
            "tab_wf",
            "tab_mmk",
            "tab_report",
        ]
    }
    return copy.deepcopy(config)


def get_phase_color(num):
    cols = config["plt"]["phase_colors"]
    while num > len(cols):
        num -= len(cols)  # periodic colors
    return cols[num]


def get_line_color(num):
    cols = config["plt"]["line_colors"]
    while num >= len(cols):
        num -= len(cols)  # periodic colors
    return cols[num]


def get_config(default=False):
    """
    Returns the configuration dictionary
    
    Parameters
    ----------
    default : Boolean
              if True:  returns the default config
              if False: returns the config from the home-directory of the user
             
    Returns
    -------
    return : dictionary
             config dict
    """
    if default:
        return get_init_config()
    else:
        config = load_config()
        if config:
            return config
        else:
            return get_init_config()


def save_config(config):
    """
    Saves the configuration in the home directory of the user
    
    Parameters
    ----------
    config : Dictionary
             config dict
    """
    user_config_dir = get_config_dir()
    if not os.path.isdir(user_config_dir):
        os.makedirs(user_config_dir)
    with open(os.path.join(user_config_dir, "config.json"), "w") as f:
        json.dump(config, f, indent=2)


def load_config():
    """
    Saves the configuration in the home directory of the user
    
    Returns
    -------
    return : dictionary
             config dict from user home directory. Use default config
             if there is no config file of the user
    """
    user_config_dir = get_config_dir()
    if not os.path.isdir(user_config_dir):
        os.makedirs(user_config_dir)

    if os.path.isfile(os.path.join(user_config_dir, "config.json")):
        with open(os.path.join(user_config_dir, "config.json")) as f:
            config = json.load(f)
            # test for new data
            config_init = get_init_config()
            for key, value in config_init.items():
                if key not in config.keys():
                    config[key] = value
            for key, value in config_init["plt"].items():
                if key not in config["plt"].keys():
                    config["plt"][key] = value
    else:
        return get_init_config()
    return config


config = get_config()



# -*- coding: utf-8 -*-

import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import numpy as np
from swat_em.datamodel import datamodel


def test_is_symmetric():
    wdg = datamodel()
    wdg.genwdg(Q=12, P=2, m=3, w=5, layers=2)
    assert wdg.get_is_symmetric() == True


def test_fundamental_winding_factor():
    wdg = datamodel()
    wdg.genwdg(Q=12, P=2, m=3, w=5, layers=2)
    wf = wdg.get_fundamental_windingfactor()
    wf_expected = [0.933013, 0.933013, 0.933013]
    np.testing.assert_allclose(wf, wf_expected, rtol=1e-3, atol=0)


def test_num_series_turns():
    wdg = datamodel()
    wdg.genwdg(Q=12, P=2, m=3, w=5, layers=2)
    assert wdg.get_num_series_turns() == 4


def test_radial_force_modes():
    wdg = datamodel()
    wdg.genwdg(Q=12, P=2, m=3, w=5, layers=2)
    assert wdg.get_radial_force_modes() == [2, 4, 6]


def test_periodicity():
    wdg = datamodel()
    wdg.genwdg(Q=12, P=2, m=3, w=5, layers=2)
    assert wdg.get_periodicity_t() == 1


def test_parallel_connections():
    wdg = datamodel()
    wdg.genwdg(Q=12, P=2, m=3, w=5, layers=2)
    assert wdg.get_parallel_connections() == [1, 2]


def test_lcmQP():
    wdg = datamodel()
    wdg.genwdg(Q=12, P=2, m=3, w=5, layers=2)
    assert wdg.get_lcmQP() == 12


if __name__ == "__main__":
    test_is_symmetric()
    test_fundamental_winding_factor()
    test_num_series_turns()
    test_radial_force_modes()
    test_periodicity()
    test_parallel_connections()
    test_lcmQP()



# -*- coding: utf-8 -*-

import os
import shutil
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import swat_em
from swat_em.datamodel import datamodel
from swat_em.report import HtmlReport


def clean_up():
    for f in ["./report.html", "export.txt"]:
        try:
            os.remove(f)
        except:
            pass
    try:
        shutil.rmtree("report-files/")
    except:
        pass


def test_html_report():
    clean_up()
    data = datamodel()
    data.genwdg(12, 10, m=3, w=1, layers=2, turns=1)
    data.set_title("12-10 machine")
    data.set_notes("This is an example machine")

    rep = HtmlReport(data)
    #  rep.create()
    #  rep.open_in_browser()

    fname = rep.create("./report.html")
    print("filename:", fname)
    assert os.path.isfile("./report.html")
    clean_up()


def test_text_report():
    clean_up()
    data = datamodel()
    data.genwdg(12, 10, m=3, w=1, layers=2, turns=1)
    data.set_title("12-10 machine")
    data.set_notes("This is an example machine")
    data.export_text_report("export.txt")
    assert os.path.isfile("export.txt")
    clean_up()


if __name__ == "__main__":
    test_html_report()
    test_text_report()



# -*- coding: utf-8 -*-
# testcase for individual number of turns

# winding data taken from
# AC-Motoren mit verteilter Zahnspulenwicklung
# Gerhard Huth, Jens Krotsch
# e & i Elektrotechnik und Informationstechnik 2019
# https://dx.doi.org/10.1007/s00502-019-0717-9

import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from swat_em.datamodel import datamodel
import swat_em.analyse
import numpy as np


def test_doza():
    Q = 12
    p = 1
    m = 3
    wstep = 1
    layers = 2
    U = [[1, 2, -11, -12], [5, 6, -7, -8]]
    V = [[-3, -4, 5, 6], [9, 10, -11, -12]]
    W = [[-7, -8, 9, 10], [1, 2, -3, -4]]
    S = [U, V, W]

    zs = 1
    dz = [zs * 0.0, zs * 0.1, zs * 0.2, zs * 0.3]
    kw1 = [0.4830, 0.5054, 0.5278, 0.5502]
    kw5 = [-0.1294, -0.0458, -0.0379, -0.1215]
    kw7 = [0.1294, 0.0458, 0.0379, 0.1215]

    for k in range(len(dz)):
        za = zs + dz[k]
        zb = zs - dz[k]
        turns = [
            [[zb, za, za, zb], [za, zb, zb, za]],
            [[za, zb, zb, za], [za, zb, zb, za]],
            [[za, zb, zb, za], [za, zb, zb, za]],
        ]

        data = datamodel()
        data.set_machinedata(Q=Q, p=p, m=m)
        data.set_phases(S, turns=turns)
        data.analyse_wdg()
        #  data.save_to_file('doza_{}.wdg'.format(dz[k]))
        bc, txt = data.get_basic_characteristics()
        #  print(bc)

        idx = data.results["nu_el"].index(1)
        assert kw1[k] == np.round(data.results["kw_el"][idx][0], 4)  # phase U
        assert kw1[k] == np.round(data.results["kw_el"][idx][1], 4)  # phase V
        assert kw1[k] == np.round(data.results["kw_el"][idx][2], 4)  # phase W
        idx = data.results["nu_el"].index(5)
        assert kw5[k] == np.round(data.results["kw_el"][idx][0], 4)  # phase U
        assert kw5[k] == np.round(data.results["kw_el"][idx][1], 4)  # phase V
        assert kw5[k] == np.round(data.results["kw_el"][idx][2], 4)  # phase W
        idx = data.results["nu_el"].index(7)
        assert kw7[k] == np.round(data.results["kw_el"][idx][0], 4)  # phase U
        assert kw7[k] == np.round(data.results["kw_el"][idx][1], 4)  # phase V
        assert kw7[k] == np.round(data.results["kw_el"][idx][2], 4)  # phase W


if __name__ == "__main__":
    test_doza()



# -*- coding: utf-8 -*-

import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import numpy as np
from swat_em.datamodel import datamodel
from swat_em.config import config

# The double linked leakage depends on the infinity fourier series
# of the mmf. The number of harmonic numbers depends on the number of
# points in the time domain. For the tests these number of points
# are set to a fixed value:
config["num_MMF_points"] = 3601


def test1():
    print("tooth coil winding 12/10")
    wdg = datamodel()
    wdg.genwdg(Q=12, P=10, m=3, w=1, layers=2)
    sigma_d = wdg.get_double_linked_leakage()
    #  print('sigma_d:', sigma_d)
    np.testing.assert_allclose(sigma_d, 0.968, rtol=1e-3, atol=0)


def test2():
    print("tooth coil winding 12/14")
    wdg = datamodel()
    wdg.genwdg(Q=12, P=14, m=3, w=1, layers=2)
    sigma_d = wdg.get_double_linked_leakage()
    #  print('sigma_d:', sigma_d)
    np.testing.assert_allclose(sigma_d, 2.8579, rtol=1e-3, atol=0)


def test3():
    print("single layer winding with constant number of turns")
    wdg = datamodel()
    wdg.genwdg(Q=6, P=2, m=3, w=-1, layers=1)
    sigma_d = wdg.get_double_linked_leakage()
    #  print('sigma_d:', sigma_d)
    np.testing.assert_allclose(sigma_d, 0.09662, rtol=1e-3, atol=0)


def test4():
    print("single layer winding with constant number of turns")
    wdg = datamodel()
    turns = 1
    S = [[1, -4], [-2, 5], [3, -6]]
    turns = [[1, 1], [1, 1], [1, 1]]
    wdg.set_machinedata(Q=6, p=1, m=3)
    wdg.set_phases(S, turns=turns)
    wdg.analyse_wdg()
    sigma_d = wdg.get_double_linked_leakage()
    #  print('sigma_d:', sigma_d)
    np.testing.assert_allclose(sigma_d, 0.09662, rtol=1e-3, atol=0)


def test5():
    print("single layer winding with constant number of turns")
    wdg = datamodel()
    turns = 100
    S = [[1, -4], [-2, 5], [3, -6]]
    turns = [[1, 1], [1, 1], [1, 1]]
    wdg.set_machinedata(Q=6, p=1, m=3)
    wdg.set_phases(S, turns=turns)
    wdg.analyse_wdg()
    sigma_d = wdg.get_double_linked_leakage()
    #  print('sigma_d:', sigma_d)
    np.testing.assert_allclose(sigma_d, 0.09662, rtol=1e-3, atol=0)


def test6():
    print("single layer winding with variable number of turns")
    wdg = datamodel()
    turns = [[[10, 10], []], [[10, 10], []], [[10, 10], []]]
    S = [[1, -4], [-2, 5], [3, -6]]
    turns = [[1, 1], [1, 1], [1, 1]]
    wdg.set_machinedata(Q=6, p=1, m=3)
    wdg.set_phases(S, turns=turns)
    wdg.analyse_wdg()
    sigma_d = wdg.get_double_linked_leakage()
    #  print('sigma_d:', sigma_d)
    np.testing.assert_allclose(sigma_d, 0.09662, rtol=1e-3, atol=0)


if __name__ == "__main__":
    test1()
    test2()
    test3()
    test4()
    test5()
    test6()



# -*- coding: utf-8 -*-
"""
Provides a central class for data 
"""
import json
import os
import fractions
import math
import copy
import string
import gzip
import gc
import numpy as np
from swat_em import analyse
from swat_em import report as rep
from swat_em import wdggenerator
from swat_em.config import config, get_phase_color
from swat_em import plots


class datamodel:
    """
    Provides a central place for all data. All analysis functions are
    connect with this class.
    """

    file_format = 1
    machinedata_keys = ["Q", "p", "m", "phases", "wstep", "Qes"]
    results_keys = [
        "q",
        "nu_el",
        "Ei_el",
        "kw_el",
        "phaseangle_el",
        "nu_mech",
        "Ei_mech",
        "kw_mech",
        "phaseangle_mech",
        "valid",
        "error",
    ]

    def __init__(self):
        self.reset_data()
        self.reset_results()

    def __str__(self):
        txt = []
        txt.append("WINDING DATAMODEL")
        txt.append("=================")
        txt.append("")
        if self.results["kw_el"] != None:
            txt.append("Title: " + self.get_title())
            txt.append("Number of slots:  {}".format(self.get_num_slots()))
            txt.append("Number of poles:  {}".format(2 * self.get_num_polepairs()))
            txt.append("Number of phases: {}".format(self.get_num_phases()))
            txt.append("Number of layers: {}".format(self.get_num_layers()))
            txt.append("Winding step    : {}".format(self.get_windingstep()))

            txt.append("Number of slots per pole per phase: {}".format(self.get_q()))
            wf = [str(round(k, 3)) for k in self.get_fundamental_windingfactor()]
            txt.append("Fundamental winding factor: {}".format(", ".join(wf)))
        else:
            txt.append("-- NOW WINDING DEFINED IN THIS MODEL --")
        return "\n".join(txt)

    def reset_data(self):
        """
        resets all data of the datamodel
        """
        self.title = ""
        self.notes = ""
        self.machinedata = {}
        for key in self.machinedata_keys:
            self.machinedata[key] = None
        self.set_turns(1)
        self.set_machinedata(Qes=0)
        self.generator_info = {}

    def reset_results(self):
        """
        Remove all existing results
        """
        self.results = {}
        for key in self.results_keys:
            self.results[key] = None

    def set_title(self, title):
        """
        Set the title of the winding 

        Parameters
        ----------
        title : string
                title
        """
        self.title = title

    def get_title(self):
        """
        Get the title of the winding 

        Returns
        ----------
        title : string
                title
        """
        if self.title == "":
            return "Untitled"
        else:
            return self.title

    def get_notes(self):
        """
        Get notes for the winding 

        Returns
        ----------
        notes : string
                Some notes
        """
        return self.notes

    def set_notes(self, notes):
        """
        Set additional notes for the winding 

        Parameters
        ----------
        notes :  string
                 Some notes
        """
        self.notes = notes

    def set_machinedata(self, Q=None, p=None, m=None, Qes=None):
        """
        setting the machine data 

        Parameters
        ----------
        Q :      integer
                 number of slots
        p :      integer
                 number of pole pairs
        m:       integer
                 number of phases
        """
        if Q:
            self.set_num_slots(int(Q))
        if p:
            self.set_num_polepairs(int(p))
        if m:
            self.set_num_phases(int(m))
        if Qes:
            self.set_num_empty_slots(int(Qes))

    def set_phases(self, S, turns=1, wstep=None):
        """
        setting the winding layout
        
        Parameters
        ----------
        S : list of lists
            winding layout for every phase, for example:
            S = [[1,-2], [3,-4], [5,-6]]. This means there are 3 phases
            with phase 1 in in slot 1 and in slot 2 with negativ winding direction. 
            For double layer windings there must be additional lists:
            S = [[[1, -4], [-3, 6]], [[3, -6], [-5, 2]], [[-2, 5], [4, -1]]]
            Hint: [[[first layer], [second layer]], ... ]
                     
        wstep : integer
                winding step (slots as unit)
        """
        if not hasattr(S[0][0], "__iter__"):
            for k in range(len(S)):
                S[k] = [S[k], []]
        if hasattr(turns, "__iter__"):
            if not hasattr(turns[0][0], "__iter__"):
                for k in range(len(turns)):
                    turns[k] = [turns[k], []]

        self.machinedata["phases"] = S
        self.set_turns(turns)
        self.set_machinedata(m=len(S))
        self.machinedata["phasenames"] = [
            string.ascii_uppercase[k] for k in range(len(S))
        ]
        if wstep:
            self.set_windingstep(wstep)

    def set_valid(self, valid, error, info=""):
        self.generator_info["valid"] = valid
        self.generator_info["error"] = error
        self.generator_info["info"] = info

    def genwdg(self, Q, P, m, layers, w=-1, turns=1, empty_slots=0, analyse=True):
        """
        Generates a winding layout and stores it in the datamodel

        Parameters
        ----------
        Q :      integer
                 number of slots
        P :      integer
                 number of poles
        m :      integer
                 number of phases
        w :      integer
                 winding step (1 for tooth coils)
        layers : integer
                 number of coil sides per slot    
        turns  : integer
                 number of turns per coil
        analyse : Bool
                  If False the generated winding doesn't get analysed
        empty_slots : integer
                    Defines the number of empty slots ("dead coil winding")
                    -1: Choose number of empty slots automatically (the smallest
                      possible number is choosen)
                    0: No empty slots
                    >0: Manual defined number of empty slots
        """

        wdglayout = wdggenerator.genwdg(Q, P, m, w, layers, empty_slots)
        if wdglayout is None:
            return
        self.set_machinedata(Q, int(P / 2), m, wdglayout["Qes"])

        self.set_phases(S=wdglayout["phases"], turns=turns, wstep=wdglayout["wstep"])
        self.set_valid(
            valid=wdglayout["valid"], error=wdglayout["error"], info=wdglayout["info"]
        )
        if analyse:
            self.analyse_wdg()

    def get_num_layers(self):
        """
        Returns the number of layers of the actual winding layout
        """
        l = 1
        for p in self.get_phases():
            if len(p[1]) > 0:
                l = 2
        return l

    def get_basic_characteristics(self):
        """
        Returns the basic charactericits of the winding as 
        dictionary and a html string
        """
        if not "basic_char" in self.results.keys():
            bc = analyse.get_basic_characteristics(
                self.get_num_slots(),
                2 * self.get_num_polepairs(),
                self.get_num_phases(),
                self.get_phases(),
                self.get_turns(),
                self.get_num_empty_slots(),
            )
            bc["r"] = self.get_radial_force_modes(
                num_modes=config["radial_force"]["num_modes"]
            )
            bc["sigma_d"] = self.get_double_linked_leakage()
            bc["t"] = self.get_periodicity_t()
            self.results["basic_char"] = bc
        else:
            bc = self.results["basic_char"]

        dat = [
            ["Number of slots ", rep.italic("Q: "), str(self.get_num_slots())],
            [
                "Number of empty slots ",
                rep.italic("Qes: "),
                str(self.get_num_empty_slots()),
            ],
            ["Number of poles ", rep.italic("2p: "), str(2 * self.get_num_polepairs())],
            ["Number of phases ", rep.italic("m: "), str(self.get_num_phases())],
            ["slots per 2p per m ", rep.italic("q: "), str(bc["q"])],
            ["winding step ", rep.italic("ws: "), str(self.get_windingstep())],
        ]

        for i, k in enumerate(bc["kw1"]):
            dat.append(
                [
                    "winding factor (m={}) ".format(i + 1),
                    rep.italic("kw1: "),
                    str(round(k, 3)),
                ]
            )
        dat.append(
            [
                "double linked leakage ",
                rep.italic("σ<sub>d: </sub>"),
                str(round(bc["sigma_d"], 3)),
            ]
        )
        dat.append(["lcm(Q, P) ", "", str(bc["lcmQP"])])
        dat.append(["periodic base winding ", rep.italic("t: "), str(bc["t"])])

        a_ = [str(i) for i in analyse.Divisors(bc["a"])]
        dat.append(["parallel connection ", rep.italic("a: "), ",".join(a_)])
        r_ = [str(i) for i in bc["r"]]
        dat.append(["radial force modes ", rep.italic("r: "), ",".join(r_) + ",.."])
        if bc["sym"]:  # and bc['a']:
            dat.append(["symmetric ", "", str(bc["sym"])])
        else:
            dat.append([rep.red("symmetric "), "", rep.red(str(bc["sym"]))])
        #  if 'valid' in self.generator_info.keys():
        #  if self.generator_info['valid']:
        #  dat.append(['valid ', '', 'True'])
        #  else:
        #  dat.append([rep.red('valid '), '', rep.red('False')])

        txt = rep.table(dat)
        if bc["error"]:
            txt.append(rep.red("error: " + str(bc["error"])))

        if len(self.notes) > 0:
            #  txt.append('<br><br>' + rep.bold(Notes) + ':<br>')
            txt.append(rep.bold("<br><br>Notes<br>"))
            txt.append(rep.italic(self.notes))

        txt = "".join(txt)
        return bc, txt

    def get_radial_force_modes(self, num_modes=None):
        """
        Returns the radial force modes caused by the winding.
        The results includes also the modes with a multiple of the 
        phase-number (which aren't there if the machine is star-connected). 

        Parameters
        ----------
        num_modes : integer
                    Max. number of modes. If not given the default value
                    from the config file is used

        Returns
        -------
        MMK: list
             radial force modes
        """
        if num_modes == None:
            num_modes = config["radial_force"]["num_modes"]
        if "MMK" not in self.results.keys():
            self._calc_MMK()
        return analyse.calc_radial_force_modes(
            self.results["MMK"]["MMK"], self.get_num_phases(), num_modes=num_modes
        )

    def get_num_series_turns(self):
        """
        Returns the number of turns in series per phase.
        If the number of coil sides per phase or number of turns per
        phase is not identically than a mean value of turns of all
        phases is returned.

        Returns
        -------
        w: number
           number of turns in series per phase
        """
        S2 = analyse._flatten(self.get_phases())
        turns2 = self.get_turns()
        if hasattr(turns2, "__iter__"):
            turns2 = analyse._flatten(turns2)
        if not hasattr(turns2, "__iter__"):  # constant number of turns
            w = 0
            for k in S2:
                w += len(k)
            w *= turns2
        else:  # variable number of turns
            w = 0
            for k1 in range(len(S2)):
                for k2 in range(len(S2[k1])):
                    w += S2[k1][k2] * turns2[k1][k2]
        w = w / 2 / self.get_num_phases()
        return w

    def get_MMF_harmonics(self, threshold=False):
        """
        Returns the harmonics of the MMF curve 
        
        Parameters
        ----------
        threshold: float
                   Relative threshold for cutting off small amplit>udes
                   0 < threshold <= 1.0
               
        Returns
        -------
        nu: 1D numpy array
            ordinal number (mechanical)
        Cnu: 1D numpy array
             amplitude corresponding to nu
        phase: 1D numpy array
               phaseangle of the harmonic corresponding to nu in
               range between -pi and +pi
        """
        if "MMK" not in self.results.keys():
            self._calc_MMK()
        nu = np.array(self.results["MMK"]["nu"])
        Cnu = np.abs(self.results["MMK"]["HA"])
        phase = np.angle(self.results["MMK"]["HA"])
        if threshold:
            idx = Cnu > np.max(Cnu) * threshold
            nu = nu[idx]
            Cnu = Cnu[idx]
            phase = phase[idx]
        return nu, Cnu, phase

    def get_double_linked_leakage(self):
        """
        Returns the coefficient of the double linkead leakage flux.
        This number is a measure of the harmonic content of the 
        MMF in the airgap caused by the winding. As higher the number
        as higher the harmonics.

        Returns
        -------
        sigma_d: float
                 coefficient of the double linkead leakage flux
        """
        nu, Cnu, _ = self.get_MMF_harmonics()
        if nu[0] == 0:
            nu = nu[1:]
            Cnu = Cnu[1:]
        p = self.get_num_polepairs()
        I = 1  # current for MMK
        w = self.get_num_series_turns()
        if w != 0:
            kw = Cnu * np.pi * nu / (3 * np.sqrt(2) * I * w)
            sigma_d = analyse.double_linked_leakage(kw, nu, p)
        else:
            sigma_d = -1
        return sigma_d

    def calc_num_basic_windings_t(self):
        """
        Calculates and returns the number of basic windings 't' for the actual
        winding layout

        Returns
        -------
        t: integer
           Periodicity for the winding layout
        """
        l, ls, lcol = self.get_layers()
        layers, Q = l.shape

        t = 1
        i = 2
        while i <= Q // 2:
            if Q % i == 0:
                length = Q // i
                is_t = True
                for k in range(i - 1):
                    if not np.all(
                        l[:, k * length : (k + 1) * length]
                        == l[:, (k + 1) * length : (k + 2) * length]
                    ):
                        is_t = False
                if is_t:
                    t = i
            i += 1
        return t

    def get_num_slots(self):
        """
        Returns the number of slots Q

        Returns
        -------
        Q: integer
           number of slots
        """
        return self.machinedata["Q"]

    def set_num_slots(self, Q):
        """
        Sets the number of slots Q 

        Parameters
        ----------
        Q : integer
            number of slots
        """
        self.machinedata["Q"] = Q

    def get_num_polepairs(self):
        """
        Returns the number of pole-pairs p

        Returns
        -------
        p: integer
           number of pole-pairs
        """
        return self.machinedata["p"]

    def set_num_polepairs(self, p):
        """
        Sets the number of pole pairs p 

        Parameters
        ----------
        p : integer
            number of pole pairs
        """
        self.machinedata["p"] = p

    def get_num_phases(self):
        """
        Returns the number of phases m 

        Returns
        -------
        m: integer
           number of phases
        """
        return self.machinedata["m"]

    def set_num_phases(self, m):
        """
        Sets the number of phases m 

        Parameters
        ----------
        m : integer
            number of phases
        """
        self.machinedata["m"] = m

    def get_windingstep(self):
        """
        Returns the winding step 

        Returns
        -------
        w: integer
           winding step
        """
        return self.machinedata["wstep"]

    def set_windingstep(self, w):
        """
        Sets the winding step w

        Parameters
        ----------
        w : integer
            winding step
        """
        self.machinedata["wstep"] = w

    def set_num_empty_slots(self, Qes):
        self.machinedata["Qes"] = Qes

    def get_num_empty_slots(self):
        if self.machinedata["Qes"] is not None:
            return self.machinedata["Qes"]
        else:
            return 0

    def get_phases(self, flatten=False):
        """
        Returns the definition of the winding layout. For every phase
        there is a sublist which contains the slot number which are 
        allocated to the phase.
        phases

        phases[0] contains the slot numbers for the first phase
        phases[1] contains the slot numbers for the second phase
        phases[m-1] contains the slot numbers for the last phase

        Returns
        -------
        phases: list of lists
                winding layout
        """
        if flatten:
            return analyse._flatten(self.machinedata["phases"])
        else:
            return self.machinedata["phases"]

    def get_layers(self):
        """
        Returns the definition of the winding layout alternative to the
        'get_phases' function. For every layer (with the length of the
        number of slots) there is a sublist which contains the phase-number.
        
        layers[0][0] contains the phase number for first layer and first slot
        layers[0][1] contains the phase number for first layer and second slot
        layers[0][Q-1] contains the phase number for first layer and last slot
        layers[1][0] contains the phase number for second layer and first slot

        Returns
        -------
        layers:     numpy array 
                    winding layout
        slayers:    numpy array 
                    same as 'layers' but as string
        layers_col: numpy array 
                    phase colors
        """
        N = self.get_num_layers()
        Q = self.get_num_slots()
        layers = np.zeros([N, Q], dtype=int)
        layers_s = np.array(layers, dtype=str)
        layers_col = np.array(layers, dtype=str)
        layers_col[:] = "#FFFFFF"
        m = self.get_num_phases()
        for kl in range(N):
            for km in range(m):
                col = get_phase_color(km)
                for kcs in range(len(self.machinedata["phases"][km][kl])):
                    slot = self.machinedata["phases"][km][kl][kcs]
                    if slot > 0:
                        layers[kl, abs(slot) - 1] = km + 1
                        layers_s[kl, abs(slot) - 1] = "+" + str(km + 1)
                    elif slot < 0:
                        layers[kl, abs(slot) - 1] = -(km + 1)
                        layers_s[kl, abs(slot) - 1] = "-" + str(km + 1)
                    layers_col[kl, abs(slot) - 1] = col
        return layers, layers_s, layers_col

    def get_phasenames(self):
        """
        Returns the names of the phases as a series of characters
        'A', 'B', 'C', ... with length of the number of phases
        
        Returns
        -------
        phasenames: list
                    names of the phases

        Examples
        --------
        if there are m = 3 phases:
        >>> data.get_phasenames()
        ['A', 'B', 'C']
        """
        return self.machinedata["phasenames"]

    def get_windingfactor_el(self):
        """
        Returns the windings factors with respect to the electrical
        ordinal numbers
        
        Returns
        -------
        nu: numpy array
            ordinal numbers
        kw: 2D numpy array
            windings factors, (one column for each phase)
        """
        return np.array(self.results["nu_el"]), np.array(self.results["kw_el"])

    def get_windingfactor_el_by_nu(self, nu):
        """
        Returns the windings factors with respect to the electrical
        ordinal numbers for the given ordinal number
        
        Parameters
        ----------
        nu: integer
            ordinal number 
               
        Returns
        -------
        kw: 1D numpy array
            windings factors, (one column for each phase)
        """
        kw = analyse.calc_kw_by_nu(
            self.get_num_slots(),
            self.get_phases(),
            self.get_turns(),
            self.get_num_polepairs(),
            nu,
        )
        return kw

    def get_windingfactor_mech(self):
        """
        Returns the windings factors with respect to the mechanicl
        ordinal numbers
        
        Returns
        -------
        nu: numpy array
            ordinal numbers
        kw: 2D numpy array
            windings factors, (one column for each phase)
        """
        return np.array(self.results["nu_mech"]), np.array(self.results["kw_mech"])

    def get_windingfactor_mech_by_nu(self, nu):
        """
        Returns the windings factors with respect to the mechanical
        ordinal numbers for the given ordinal number
        
        Parameters
        ----------
        nu: integer
            ordinal number 
               
        Returns
        -------
        kw: 1D numpy array
            windings factors, (one column for each phase)
        """
        kw = analyse.calc_kw_by_nu(
            self.get_num_slots(), self.get_phases(), self.get_turns(), 1, nu
        )
        return kw

    def get_fundamental_windingfactor(self):
        """
        Returns the fundamental winding factors for each phase
        
        Returns
        -------
        kw: list
            windings factors, (one entry for each phase)
        """
        wf = self.get_windingfactor_el_by_nu(1)
        if wf is None:
            return None
        else:
            return [abs(kw) for kw in wf]
        #  return self.results['kw_el'][0]

    def get_turns(self):
        """
        Returns the number of turns. If all coil sides has the same
        number of turns, the return value is a scalar. If every coil
        side has an individual number of turns, the return value
        consists of lists with the same shape as the winding layout
        (phases)
        
        Returns
        -------
        turns: integer, float or list of lists
               number of turns 
        """
        return self.machinedata["turns"]

    def get_is_symmetric(self):
        """
        Returns the symmetry of the winding
        
        Returns
        -------
        is_symmetric: Boolean
                      True if the winding is symmetric
                      False if the winding is not symmetric 
        """
        return self.results["wdg_is_symmetric"]

    def get_periodicity_t(self):
        """
        Returns the periodicity of the winding. In most cases 
        t = gcd(Q,p). For user defined windings this may be different.
        
        Returns
        -------
        t: integer
           Number of periodic base windings
        """
        if "t" not in self.results:
            try:
                self.results["t"] = self.calc_num_basic_windings_t()
            except:
                self.results["t"] = -1
        return self.results["t"]

    def get_parallel_connections(self):
        """
        Returns all possible parallel connections of the winding.
        
        Returns
        -------
        a: list
           Number of possible parallel connections
        """
        return [i for i in analyse.Divisors(self.results["a"])]

    def get_lcmQP(self):
        """
        Returns the lowest common multiple of the slot number Q and
        the number of Poles. For permanent-magnet machines this value
        represents the first ordinal number for the cogging torque.
        
        Returns
        -------
        lcmQP: integer
               Lowest common multiple lcm(Q, P)
        """
        return self.results["lcmQP"]

    def set_turns(self, turns):
        """
        Sets the number of turns. If all coil sides has the same
        number of turns, the parameter should be an scalar. If every coil
        side has an individual number of turns, the parameter value
        have to consist of lists with the same shape as the winding layout
        (phases)
        
        Parameters
        ----------
        turns: integer, float or list of lists
               number of turns 
        """
        self.machinedata["turns"] = turns

    def get_q(self):
        """
        Returns the number of slots per pole per phase.

        Returns
        -------
        layers: Fraction
                number of slots per pole per phase
        """
        if "q" not in self.results.keys():
            q = fractions.Fraction(
                (self.get_num_slots() - self.get_num_empty_slots())
                / (self.get_num_phases() * 2 * self.get_num_polepairs())
            ).limit_denominator(100)
            self.results["q"] = q
        return self.results["q"]

    def _set_q(self, q):
        """
        Sets the number of slots per pole per phase q 

        Parameters
        ----------
        q : integer
            number of slots per pole per phase
        """
        self.results["q"] = q

    def analyse_wdg(self):
        """
        Do a detailled analyses of the winding. This includes
        winding factors, detection of periodicity and symmetry, 
        radial force modes and so on. Use the get_* functions for 
        getting the results.
        """
        self.reset_results()

        # number of slots per pole and phase
        q = fractions.Fraction(
            (self.get_num_slots() - self.get_num_empty_slots())
            / (self.get_num_phases() * 2 * self.get_num_polepairs())
        ).limit_denominator(100)
        self._set_q(q)

        # base winding
        #  self.results['t'] = math.gcd(self.machinedata['Q'], self.machinedata['p'])
        self.results["t"] = self.calc_num_basic_windings_t()

        # electrical winding factor
        a, b, c, d = analyse.calc_kw(
            self.machinedata["Q"],
            self.machinedata["phases"],
            self.machinedata["turns"],
            self.machinedata["p"],
            config["N_nu_el"],
            config,
        )
        self.results["nu_el"] = a
        self.results["Ei_el"] = b
        self.results["kw_el"] = c
        self.results["phaseangle_el"] = d

        # mechanical winding factor
        a, b, c, d = analyse.calc_kw(
            self.machinedata["Q"],
            self.machinedata["phases"],
            self.machinedata["turns"],
            1.0,
            config["N_nu_mech"],
            config,
        )
        self.results["nu_mech"] = a
        self.results["Ei_mech"] = b
        self.results["kw_mech"] = c
        self.results["phaseangle_mech"] = d

        # winding symmetric?
        self.results["wdg_is_symmetric"] = analyse.wdg_is_symmetric(
            self.results["Ei_el"], self.machinedata["m"]
        )

        # periodicity of the winding (radial force, parallel connection)?
        self.results["wdg_periodic"] = analyse.wdg_get_periodic(
            self.results["Ei_el"], self.machinedata["phases"]
        )

        # MMK
        if "MMK" not in self.results.keys():
            self._calc_MMK()

        bc, bc_txt = self.get_basic_characteristics()
        self.results["basic_char"] = bc
        self.results["t"] = bc["t"]
        self.results["a"] = bc["a"]
        self.results["lcmQP"] = bc["lcmQP"]

    def _calc_MMK(self):
        phi, MMK, theta = analyse.calc_MMK(
            self.get_num_slots(),
            self.get_num_phases(),
            self.get_phases(),
            self.get_turns(),
            N=config["num_MMF_points"],
        )
        HA = analyse.DFT(MMK[:-1])
        nu = list(range(len(HA)))

        self.results["MMK"] = {}
        self.results["MMK"]["MMK"] = MMK
        self.results["MMK"]["phi"] = phi
        self.results["MMK"]["theta"] = theta
        self.results["MMK"]["nu"] = nu
        self.results["MMK"]["HA"] = HA

    def plot_layout(self, filename, res=None, show=False):
        """
        Generates a figure of the winding layout
        
        Parameters
        -------
        filename: string
                  file-name with extension to save the figure
        res: list 
             Resolution for the figure in pixes for x and y direction
             example: res = [800, 600]
        show: Bool
              If true the window pops up for interactive usage
        """
        if res == None:
            res = config["plt"]["res"]
        plt = plots._slot_plot(None, None, self)
        plt.plot_slots(self.get_num_slots())
        plt.plot(self, show=show)
        plt.save(fname=filename, res=res)

    def plot_overhang(self, filename, res=None, show=False, optimize_overhang=False):
        """
        Generates a figure of the winding layout
        
        Parameters
        -------
        filename: string
                  file-name with extension to save the figure
        res: list 
             Resolution for the figure in pixes for x and y direction
             example: res = [800, 600]
        show: Bool
              If true the window pops up for interactive usage
        optimize_overhang: Bool
              If true swat-em tries to identify shorter winding overhangs
        """
        if res == None:
            res = config["plt"]["res"]
        plt = plots._overhang_plot(None, None, self)
        plt.plot(show=show, optimize_overhang=optimize_overhang)
        plt.save(fname=filename, res=res)

    def plot_star(self, filename, res=None, ForceX=True, show=False):
        """
        Generates a figure of the star voltage phasors
        
        Parameters
        -------
        filename: string
                  file-name with extension to save the figure
        res: list 
             Resolution for the figure in pixes for x and y direction
             example: res = [800, 600]
        ForceX: Bool
                If true the voltage phasors are rotated in such way, that
                the resulting phasor of the first phase matches the 
                x-axis
        show: Bool
              If true the window pops up for interactive usage
        """
        if res == None:
            res = config["plt"]["res"]
        plt = plots._slot_star(None, None, self, None)
        plt.plot(self, harmonic_idx=0, ForceX=ForceX, show=show)
        plt.save(fname=filename, res=res)

    def plot_windingfactor(self, filename, res=None, mechanical=True, show=False):
        """
        Generates a figure of the winding layout
        
        Parameters
        -------
        filename: string
                  file-name with extension to save the figure
        m: list 
             Resolution for the figure in pixes for x and y direction
             example: res = [800, 600]
        mechanical: Bool
                    If true the winding factor is plotted with respect to the
                    mechanical ordinal numbers. If false the electrical 
                    ordinal numbers are used
        show: Bool
              If true the window pops up for interactive usage
        """
        if res == None:
            res = config["plt"]["res"]
        plt = plots._windingfactor(None, None, self, None)
        plt.plot(self, mechanical=mechanical, show=show)
        plt.save(fname=filename, res=res)

    def plot_MMK(self, filename, res=None, phase=0, show=False):
        """
        Generates a figure of the winding layout
        
        Parameters
        -------
        filename: string
                  file-name with extension to save the figure
        res: list 
             Resolution for the figure in pixes for x and y direction
             example: res = [800, 600]
        phase: float
               phase angle for the current system in electical degree
               in the range 0..360°
        show: Bool
              If true the window pops up for interactive usage
        """
        if res == None:
            res = config["plt"]["res"]
        plt = plots._mmk(None, None, self, None)
        plt.plot(self, phase=phase, show=show)
        plt.save(fname=filename, res=res)

    def plot_polar_layout(
        self, filename, res=None, optimize_overhang=False, draw_poles=False, show=False
    ):
        """
        Generates a figure of the winding layout
        
        Parameters
        -------
        filename: string
                  file-name with extension to save the figure
        res: list 
             Resolution for the figure in pixes for x and y direction
             example: res = [800, 600]
        optimize_overhang: Bool
              If true swat-em tries to identify shorter winding overhangs
        draw_poles: Bool
              If true the poles of the rotor are drawn
        show: Bool
              If true the window pops up for interactive usage
        """
        if res == None:
            res = config["plt"]["res"]
        plt = plots._polar_layout_plot(None, None, self)
        plt.plot(
            self, show=show, optimize_overhang=optimize_overhang, draw_poles=draw_poles
        )
        plt.save(fname=filename, res=res)

    def save_to_file(self, fname):
        """
        Saves the data to file. 

        Parameters
        ----------
        fname :  string
                 file name
        """
        save_models_to_file(self, fname)

    def load_from_file(self, fname, idx_in_file=0):
        """
        Load data from file. 

        Parameters
        ----------
        fname :  string
                 file name
        """
        data = load_models_from_file(fname)
        if idx_in_file > len(data):
            raise Exception(
                "defined index in greater than the available number of models in the *.wdg file"
            )
        else:
            data = data[idx_in_file]
            self.machinedata = data.machinedata
            self.results = data.results

    def export_xlsx(self, fname):
        """
        Export the results to Excel xlsx file. 

        Parameters
        ----------
        fname :  string
                 file name
        """
        rep.export_xlsx(fname, self)

    def export_text_report(self, fname):
        """
        Export winding report as a text file. 

        Parameters
        ----------
        fname :  string
                 file name
        """
        txt_rep = rep.TextReport(self)
        txt_rep.save(fname)

    def export_html_report(self, fname=None):
        """
        Returns a winding report. 
        
        Parameters
        ----------
        fname :  string
                 file name for html file. If not given a file is created
                 in the temp dir of the file system (the file name ist
                 returned by this function)
                 
        Returns
        -------
        filnename : string
                    The file name of the html-file which is stored in tmp directory
        """
        html_rep = rep.HtmlReport(self)
        return html_rep.create(fname)

    def get_text_report(self):
        """
        Returns a winding report. 

        Return
        ----------
        report :  string
                  Report
        """
        txt_rep = rep.TextReport(self)
        return txt_rep.get_report()


class project:
    """
    Provides all data-objects (all winding in workspace)
    """

    file_format = 2

    def __init__(self):
        self.models = []
        self.filename = None
        self.undo_state = []
        self.redo_state = []
        self._is_saved = False

    def get_save_state(self):
        return self._is_saved

    def set_save_state(self, state):
        self._is_saved = state

    def save_undo_state(self):
        """saves the actual state of the models for undo function"""
        self.undo_state.append(copy.deepcopy([self.models, self._is_saved]))
        self.set_save_state(False)

    def save_redo_state(self):
        """saves the actual state of the models for redo function"""
        self.redo_state.append(copy.deepcopy([self.models, self._is_saved]))

    def reset_undo_state(self):
        """delete all existings undo state saves - no undo possible any more"""
        self.undo_state = []

    def reset_redo_state(self):
        """delete all existings redo state saves - no redo possible any more"""
        self.redo_state = []

    def reset_save_state(self):
        """set the save state to False for all redo and undo actions"""
        for k in range(len(self.undo_state)):
            self.undo_state[k][1] = False
        for k in range(len(self.redo_state)):
            self.redo_state[k][1] = False

    def get_num_undo_state(self):
        """return the number of undo state saves = number of possible undo opserations"""
        return len(self.undo_state)

    def get_num_redo_state(self):
        """return the number of redo state saves = number of possible redo opserations"""
        return len(self.redo_state)

    def undo(self):
        """restores the last state"""
        if self.get_num_undo_state() > 0:
            self.models, self._is_saved = self.undo_state.pop(-1)
            gc.collect()  # force to free memory

    def redo(self):
        """restores the last state"""
        if self.get_num_redo_state() > 0:
            self.models, self._is_saved = self.redo_state.pop(-1)
            gc.collect()  # force to free memory

    def set_filename(self, filename):
        """saves the filename if a the data is load from file or
        stored to a file"""
        self.filename = filename

    def get_filename(self):
        """returns the *.wdg filename if the project ist saved.
        Otherwise None is returned"""
        return self.filename

    def gen_model_name(self):
        """
        creates an unique title for a model
        """
        titles = [data.title for data in self.models]
        i = 0
        while True:
            name = "untitled" + str(i)
            if name not in titles:
                return name
            i += 1

    def add_model(self, data):
        """
        adds 'data' model to project. Generates a unique title if there
        is no title in 'data'
        """
        if data.title == "":
            name = self.gen_model_name()
            data.set_title(name)
        self.models.append(data)

    def get_titles(self):
        """returns the title of all models"""
        return [data.title for data in self.models]

    def get_num_models(self):
        """returns the number of models in this project"""
        return len(self.models)

    def get_model_by_index(self, idx):
        """returns the model with the index 'idx' """
        return self.models[idx]

    def delete_model_by_index(self, idx):
        """deletes the model with the index 'idx' """
        del self.models[idx]

    def clone_by_index(self, idx):
        """duplicates the model with the index 'idx' """
        data = copy.deepcopy(self.models[idx])
        data.set_title(data.get_title() + "_copy")
        self.add_model(data)

    def rename_by_index(self, idx, newname):
        """change the title of the model with the index 'idx' 
        with name 'newname' """
        self.models[idx].set_title(newname)

    def replace_model_by_index(self, newmodel, idx):
        """replaces the model of index 'idx' with 'newmodel' """
        self.models[idx] = newmodel

    def analyse_all_models(self):
        """analyse/recalculate all existing models"""
        for m in self.models:
            m.analyse_wdg()

    def save_to_file(self, fname):
        """
        Saves the data to file. 

        Parameters
        ----------
                 
        fname :  string
                 file name
        """
        save_models_to_file(self.models, fname, file_format=self.file_format)
        self.reset_save_state()
        self.set_save_state(True)

    def load_from_file(self, fname):
        """
        Load data from file. 

        Parameters
        ----------
        fname :  string
                 file name
        """
        models = load_models_from_file(fname)
        if len(models) > 0:
            self.models = []
            for m in models:
                self.add_model(m)
            self.set_filename(fname)
            self.reset_undo_state()
            self.reset_redo_state()
            self.reset_save_state()
            self.set_save_state(True)


def save_models_to_file(models, fname, file_format=2):
    """
    Saves winding models to file. 
    The data is stored as a json file.

    Parameters
    ----------
    models:  datamodel object or list of datamodel objects
             models to save
             
    fname :  string
             file name
    """
    if not hasattr(models, "__iter__"):
        models = [models]
    M = {}
    M["file_format"] = file_format
    M["models"] = []
    for data in models:
        N = {}
        N["machinedata"] = data.machinedata
        if type(N["machinedata"]["wstep"]) == type(fractions.Fraction()):
            N["machinedata"]["wstep"] = str(N["machinedata"]["wstep"])
        N["title"] = data.title
        N["notes"] = data.notes
        M["models"].append(N)

    # save as ASCII file
    with open(fname, "w") as f:
        json.dump(M, f, indent=2)


def load_models_from_file(fname):
    """
    Load winding models from file. 

    Parameters
    ----------
    fname :  string
             file name
    """
    try:
        if os.path.isfile(fname):
            with open(fname) as f:
                M = json.load(f)
    except:
        # file_format 1 is saved compressed
        import gzip

        with gzip.GzipFile(fname) as f:  # gzip
            s = f.read()  # bytes
            s = s.decode("utf-8")  # string
            M = json.loads(s)  # data

    if M["file_format"] == 1:
        models = []
        data = datamodel()
        data.machinedata = M["machinedata"]
        if "turns" not in data.machinedata.keys():
            data.machinedata["turns"] = 1
        data.analyse_wdg()
        models = [data]

    elif M["file_format"] == 2:
        models = []
        for m in M["models"]:
            data = datamodel()
            for key, value in m["machinedata"].items():
                data.machinedata[key] = value
            if type(data.get_windingstep()) == type(""):
                data.set_windingstep(fractions.Fraction(data.get_windingstep()))
            data.set_title(m["title"])
            data.set_notes(m["notes"])
            data.analyse_wdg()
            models.append(data)

    return models



# -*- coding: utf-8 -*-
import sys
import os
from PyQt5 import uic
from PyQt5.QtWidgets import (
    QMainWindow,
    QApplication,
    QFileDialog,
    QInputDialog,
    QMessageBox,
    QListWidgetItem,
    QMenu,
    QAction,
    QSplashScreen,
    QGraphicsScene,
)
from PyQt5.QtGui import QIntValidator, QDoubleValidator, QIcon, QPixmap, QPainter
from PyQt5.QtPrintSupport import QPrintDialog, QPrinter
from PyQt5 import QtCore

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
if getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS"):
    __dir__ = sys._MEIPASS  # for pyinstaller
else:
    __dir__ = os.path.dirname(os.path.abspath(__file__))


# Create a splash screen while loading the librarys because this
# takes some time
app = QApplication([])  # temp. app needed for splash
splash_pix = QPixmap(os.path.join(__dir__, "ui", "bitmaps", "splash.png"))
splash = QSplashScreen(splash_pix, QtCore.Qt.WindowStaysOnTopHint)
splash.show()

import time
import argparse
import subprocess
import platform
import tempfile
from swat_em import __version__
from swat_em import dialog_genwdg
from swat_em import dialog_about
from swat_em import dialog_notes
from swat_em import dialog_winding_layout
from swat_em import dialog_import_winding
from swat_em import dialog_combination_sniffer
from swat_em import dialog_settings
from swat_em.config import config, save_config, get_config
from swat_em import datamodel, project
from swat_em import wdggenerator
from swat_em import plots
from swat_em import report
from swat_em.analyse import _get_float

MSG_TIME = 3000  # time in which the message is displayed in the statusbar


class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()

        self.data = datamodel()
        self.project = project()

        # Set up the user interface from Designer.
        uic.loadUi(os.path.join(__dir__, "ui", "MainWindow.ui"), self)
        self.setWindowTitle("SWAT-EM")

        self.DIALOG_GenWinding = dialog_genwdg.GenWinding2()
        self.DIALOG_GenWindingCombinations = dialog_genwdg.GenWindingCombinations()
        self.DIALOG_CombSniffer = dialog_combination_sniffer.CombSniffer()

        # open/close tabs
        self.plot_tabs_actions = {
            "tab_slot": self.actionLayout,
            "tab_layout_polar": self.actionLayout_polar,
            "tab_overhang": self.actionOverhang,
            "tab_star": self.actionSlot_voltage_phasors,
            "tab_wf": self.actionWinding_factor,
            "tab_mmk": self.actionMagneto_motive_force,
            "tab_report": self.actionReport,
        }
        for key, value in self.plot_tabs_actions.items():
            value.triggered.connect(self.plot_tabs_update_from_menu)
        self.plot_tabs_init()
        self.actionShow_all_plots.triggered.connect(self.plot_tabs_show_all)

        # Connect the buttons
        self.Button_exit.clicked.connect(self.close)

        self.projectlist_Button_new.clicked.connect(self.dialog_new_winding)
        self.projectlist_Button_delete.clicked.connect(self.projectlist_delete)
        self.projectlist_Button_clone.clicked.connect(self.projectlist_clone)
        self.projectlist_Button_notes.clicked.connect(self.dialog_get_notes)
        self.save_report_Button.clicked.connect(self.export_to_txt)
        self.actionprint.triggered.connect(self.printer)
        self.actionFull_Screen.triggered.connect(self.toggle_fullscreen)
        self.Button_search.clicked.connect(self.find_in_report)
        self.lineEdit_find.returnPressed.connect(self.find_in_report)
        self.plot_tabs.tabCloseRequested.connect(self.plot_tabs_remove)

        # Connect menu
        self.actionExit.triggered.connect(self.close)
        self.actionNew_winding.triggered.connect(self.dialog_new_winding)
        self.actionGenerate_winding.triggered.connect(self.dialog_GenWinding)
        self.actionGenerate_winding_combinations.triggered.connect(
            self.dialog_GenWindingCombinations
        )
        self.actionCombination_sniffer.triggered.connect(self.dialog_CombSniffer)
        self.actionImport_from_file.triggered.connect(self.dialog_ImportWinding)
        self.actionAdd_Notes.triggered.connect(self.dialog_get_notes)

        self.actionsave.triggered.connect(self.save_to_file)
        self.actionsave_as.triggered.connect(self.save_as_to_file)
        self.actionload.triggered.connect(self.load_from_file)
        self.actionProgram_Info.triggered.connect(dialog_about.about)
        self.actionShow_winding_layout.triggered.connect(self.dialog_EditWindingLayout)
        self.actionSettings.triggered.connect(self.dialog_settings)
        self.actionundo.triggered.connect(self.undo)
        self.actionredo.triggered.connect(self.redo)
        self.actionFind.triggered.connect(self.set_focus_to_find)

        # Connect find in report
        self.lineEdit_find.textChanged.connect(self.find_in_report)

        # Connect project models
        self.project_listWidget.currentRowChanged.connect(self.update_project)
        self.project_listWidget.itemChanged.connect(
            self.projectlist_rename
        )  # item renamed

        # context-menu on project models
        self.project_listWidget.setContextMenuPolicy(QtCore.Qt.ActionsContextMenu)

        self.actionNew = QAction(
            "New Winding",
            self.project_listWidget,
            icon=QIcon(os.path.join(__dir__, "ui", "icons", "new.png")),
        )
        self.project_listWidget.addAction(self.actionNew)
        self.actionNew.triggered.connect(self.dialog_new_winding)

        self.actionClone = QAction(
            "clone",
            self.project_listWidget,
            icon=QIcon(os.path.join(__dir__, "ui", "icons", "clone.png")),
        )
        self.project_listWidget.addAction(self.actionClone)
        self.actionClone.triggered.connect(self.projectlist_clone)

        self.actionNotes = QAction(
            "add notes",
            self.project_listWidget,
            icon=QIcon(os.path.join(__dir__, "ui", "icons", "notes.png")),
        )
        self.project_listWidget.addAction(self.actionNotes)
        self.actionNotes.triggered.connect(self.dialog_get_notes)

        self.actionDelete = QAction(
            "delete",
            self.project_listWidget,
            icon=QIcon(os.path.join(__dir__, "ui", "icons", "delete.png")),
        )
        self.project_listWidget.addAction(self.actionDelete)
        self.actionDelete.triggered.connect(self.projectlist_delete)

        # TODO
        self.actionprint_report.triggered.connect(self.create_report)
        self.actionExport_data_to_Excel.triggered.connect(self.export_to_excel)
        self.actionExport_Text_report.triggered.connect(self.export_to_txt)

        self.actionHelp.triggered.connect(self.open_manual)

        self.plot_tabs.currentChanged.connect(
            lambda: self.update_plot_in_GUI(small_update=False)
        )
        self.radioButton_electrical.toggled.connect(self.update_plot_in_GUI)
        self.comboBox_star_harmonics.currentIndexChanged.connect(
            self.update_plot_in_GUI
        )
        self.checkBoxForceX.toggled.connect(self.update_plot_in_GUI)
        self.checkBox_opt_wdg_overhang.toggled.connect(self.update_plot_in_GUI)
        self.checkBox_opt_wdg_overhang_layout_polar.toggled.connect(
            self.update_plot_in_GUI
        )

        #  initial windings  -----------------------------------
        self.data = datamodel()
        self.data.genwdg(Q=12, P=2, m=3, w=-1, layers=1)
        self.data.set_notes(
            "Winding example for a overlapping single layer winding for a 12 slots, 2 poles machine"
        )
        self.data.set_title("12-2 overlapping winding")
        self.project.add_model(self.data)

        self.data = datamodel()
        self.data.genwdg(Q=9, P=8, m=3, w=1, layers=2, turns=10)
        self.data.set_notes(
            "Winding example for a tooth-coil-winding for 9 slots, 8 poles machine"
        )
        self.data.set_title("9-8 tooth-coil winding")
        self.project.add_model(self.data)

        self.project.set_save_state(True)  # only for the initial winding
        self.check_is_saved()

        # Plots
        self.MMK_phase_edit.setValidator(QDoubleValidator(0, 360, 1))
        self.MMK_phase_edit.setText(str(0.000))

        self.MMK_phase_slider.valueChanged.connect(self.update_MMK_phase_edit)
        self.MMK_phase_edit.textEdited.connect(self.update_MMK_phase_slider)
        self.MMK_phase_edit.textChanged.connect(
            lambda: self.update_plot_in_GUI(small_update=True)
        )

        self.fig = {}
        self.fig["tab_slot"] = plots._slot_plot(
            self.mplvl_slot, self.mplwidget_slot, self.data
        )
        self.Button_cp2clipboard_image_layout.clicked.connect(
            self.fig["tab_slot"].cp2clipboard
        )

        self.fig["tab_layout_polar"] = plots._polar_layout_plot(
            self.mplvl_layout_polar, self.mplwidget_layout_polar, self.data
        )
        self.Button_cp2clipboard_image_layout_polar.clicked.connect(
            self.fig["tab_layout_polar"].cp2clipboard
        )

        self.fig["tab_overhang"] = plots._overhang_plot(
            self.mplvl_overhang, self.mplwidget_overhang, self.data
        )
        self.Button_cp2clipboard_image_overhang.clicked.connect(
            self.fig["tab_overhang"].cp2clipboard
        )

        self.fig["tab_star"] = plots._slot_star(
            self.mplvl_star, self.mplwidget_star, self.data, self.tableWidget_star
        )
        self.Button_cp2clipboard_image_star.clicked.connect(
            self.fig["tab_star"].cp2clipboard
        )

        self.fig["tab_wf"] = plots._windingfactor(
            self.mplvl_wf, self.mplwidget_wf, self.data, self.tableWidget_wf
        )
        self.Button_cp2clipboard_image_wf.clicked.connect(
            self.fig["tab_wf"].cp2clipboard
        )

        self.fig["tab_mmk"] = plots._mmk(
            self.mplvl_mmk, self.mplwidget_mmk, self.data, self.tableWidget_mmk
        )
        self.Button_cp2clipboard_image_mmk.clicked.connect(
            self.fig["tab_mmk"].cp2clipboard
        )

        self.report = plots._report(self.reportEdit)
        self.Button_cp2clipboard_image_report.clicked.connect(self.report.cp2clipboard)
        self.highlighter = plots._Report_Highlighter(self.reportEdit)

        self.update_project_list()
        self.project.reset_undo_state()  # init-winding shouldn't undoable
        self.actionundo.setDisabled(True)
        self.actionredo.setDisabled(True)
        self.show()

    def toggle_fullscreen(self):
        if self.windowState() & QtCore.Qt.WindowFullScreen:
            self.showNormal()
        else:
            self.showFullScreen()

    def check_is_saved(self):
        """
        checks if the project ist saved for
        - setting the title of the main window with the filename
        - mark the file-name in the window title with a * if it is not saved
        - disable redu/undo buttons if there are no available states
        """
        if self.project.get_num_undo_state() < 1:
            self.actionundo.setDisabled(True)  # no more states to restore
        if self.project.get_num_redo_state() < 1:
            self.actionredo.setDisabled(True)  # no more states to restore

        fname = self.project.get_filename()
        if fname:
            if not self.project.get_save_state():
                fname = "*" + fname
            self.setWindowTitle(fname + " - SWAT-EM")

    def save_undo_state(self):
        """save the actual state of the project models for undo operation"""
        self.project.save_undo_state()
        self.actionundo.setDisabled(False)
        self.project.reset_redo_state()  # redo states aren't valid any more
        self.actionredo.setDisabled(True)
        self.check_is_saved()

    def undo(self):
        """undo the last operation (restore state)"""
        self.project.save_redo_state()
        self.actionredo.setDisabled(False)
        self.project.undo()
        self.update_project_list()
        self.check_is_saved()

    def redo(self):
        self.project.save_undo_state()
        self.actionundo.setDisabled(False)
        self.project.redo()
        self.update_project_list()
        self.check_is_saved()

    def update_project_list(self, switch_to_new=False):
        """fill the list of all data objects
        if 'switch_to_new = True: mark the last (new) object in project list 
        and show the content of it. If False: The last idx is marked again"""
        self.project_listWidget.blockSignals(True)
        idx = self.project_listWidget.currentRow()  # save current idx
        self.project_listWidget.clear()  # clear existing data
        for i, item in enumerate(self.project.get_titles()):
            widgetitem = QListWidgetItem(item)
            widgetitem.setFlags(widgetitem.flags() | QtCore.Qt.ItemIsEditable)
            self.project_listWidget.addItem(widgetitem)
        self.project_listWidget.blockSignals(False)
        if switch_to_new:
            idx = i
        else:
            idx = i if i < idx else idx
            idx = 0 if idx < 0 else idx
        self.project_listWidget.setCurrentRow(idx)  # last item is current model

    def update_project(self, idx=None):
        """update the clicked data object form the project"""
        if not idx:
            idx = self.project_listWidget.currentRow()
        self.data = self.project.get_model_by_index(idx)
        self.update_data_in_GUI()  # update all data output and figures

    def projectlist_delete(self):
        if self.project_listWidget.count() > 1:
            self.save_undo_state()  # for undo
            idx = self.project_listWidget.currentRow()
            self.project.delete_model_by_index(idx)
            self.statusbar.showMessage("winding deleted", MSG_TIME)
            self.update_project_list()

    def projectlist_clone(self):
        self.save_undo_state()  # for undo
        idx = self.project_listWidget.currentRow()
        self.project.clone_by_index(idx)
        self.statusbar.showMessage("winding cloned", MSG_TIME)
        self.update_project_list(switch_to_new=True)

    def projectlist_rename(self):
        self.save_undo_state()  # for undo
        idx = self.project_listWidget.currentRow()
        newname = self.project_listWidget.item(idx).text()
        self.project.rename_by_index(idx, newname)

    def update_MMK_phase_edit(self):
        """slider for MMK phase moved"""
        self.MMK_phase_edit.setText(str(self.MMK_phase_slider.value()))

    def update_MMK_phase_slider(self):
        """lineedit for MMK phase changed"""
        f = get_float(self.MMK_phase_edit.text())
        if f is not None:
            self.MMK_phase_slider.setValue(f)

    def dialog_GenWinding(self):
        """
        calls the dialog to generate a single winding
        """
        ret = self.DIALOG_GenWinding.run()
        if ret:
            self.save_undo_state()
            data = datamodel()
            data.genwdg(
                Q=ret["Q"],
                P=ret["P"],
                m=ret["m"],
                w=ret["w"],
                layers=ret["layers"],
                empty_slots=ret["Qes"],
            )
            if ret["overwrite"]:
                data.set_title(self.data.get_title())  # use existing title
                data.set_notes(self.data.get_notes())  # use existings notes
                self.data = data
                self.project.replace_model_by_index(
                    data, self.project_listWidget.currentRow()
                )
            else:
                self.project.add_model(data)
                self.update_project_list(switch_to_new=True)
            self.update_data_in_GUI()

    def dialog_GenWindingCombinations(self):
        """
        calls the dialog to generate a winding based on a big table of available windings
        """
        ret = self.DIALOG_GenWindingCombinations.run()
        if ret:
            self.save_undo_state()
            data = datamodel()
            data.genwdg(
                Q=ret["Q"],
                P=ret["P"],
                m=ret["m"],
                w=ret["w"],
                layers=ret["layers"],
                empty_slots=ret["Qes"],
            )

            if ret["overwrite"]:
                data.set_title(self.data.get_title())  # use existing title
                data.set_notes(self.data.get_notes())  # use existings notes
                self.data = data
                self.project.replace_model_by_index(
                    data, self.project_listWidget.currentRow()
                )
            else:
                self.project.add_model(data)
                self.update_project_list(switch_to_new=True)
            self.update_data_in_GUI()

    def dialog_CombSniffer(self):
        """
        calls the dialog to generate a winding based on a big table of available windings
        """
        ret = self.DIALOG_CombSniffer.run()
        if ret:
            self.save_undo_state()
            data = datamodel()
            data.genwdg(
                Q=ret["Q"],
                P=ret["P"],
                m=ret["m"],
                w=ret["w"],
                layers=ret["layers"],
                empty_slots=ret["Qes"],
            )

            if ret["overwrite"]:
                data.set_title(self.data.get_title())  # use existing title
                data.set_notes(self.data.get_notes())  # use existings notes
                self.data = data
                self.project.replace_model_by_index(
                    data, self.project_listWidget.currentRow()
                )
            else:
                self.project.add_model(data)
                self.update_project_list(switch_to_new=True)
            self.update_data_in_GUI()

    def dialog_ImportWinding(self):
        """Import windings from file to workspace"""
        DIALOG_ImportWinding = dialog_import_winding.import_winding()
        ret = DIALOG_ImportWinding.run()
        if ret is not None:
            self.save_undo_state()  # for undo
            for data in ret:
                self.project.add_model(data)
            self.update_project_list(switch_to_new=True)
            self.statusbar.showMessage(
                "{} winding(s) imported".format(len(ret)), MSG_TIME
            )

    def dialog_EditWindingLayout(self):
        """
        calls the dialog to generate a winding based on a big table of available windings
        """
        DIALOG_EditWindingLayout = dialog_winding_layout.layout(self.data)
        ret = DIALOG_EditWindingLayout.run()

        if ret:
            self.save_undo_state()

            data = datamodel()
            data.set_machinedata(Q=ret["Q"], p=ret["P"] // 2, m=ret["m"])
            data.set_phases(ret["phases"], wstep=ret["w"], turns=ret["turns"])
            data.analyse_wdg()
            if ret["overwrite"]:
                data.set_title(self.data.get_title())  # use existing title
                data.set_notes(self.data.get_notes())  # use existings notes
                self.data = data
                self.project.replace_model_by_index(
                    data, self.project_listWidget.currentRow()
                )
            else:
                self.project.add_model(data)
                self.update_project_list(switch_to_new=True)
            self.update_data_in_GUI()

    def dialog_get_notes(self):
        """dialog asks user to type notes and stores it to datamodel"""
        dialog = dialog_notes.get_notes(self.data.notes)
        text = dialog.run()
        if text is not None:
            self.save_undo_state()
            self.data.set_notes(text)
            self.update_data_in_GUI()

    def dialog_new_winding(self):
        """
        dialog to generate a new winding
        wraps the user input to the existing generators
        """
        dialog = dialog_genwdg.NewWinding()
        ret = dialog.run()
        if ret is not None:
            if ret == 0:  # trigger the different generators
                self.dialog_EditWindingLayout()
            elif ret == 1:
                self.dialog_GenWinding()
            elif ret == 2:
                self.dialog_GenWindingCombinations()
            else:
                raise (Exception, "winding generator {} not implemented".format(ret))

    def dialog_settings(self):
        global config
        config
        DIALOG_Settings = dialog_settings.Settings(config)
        ret = DIALOG_Settings.run()
        if ret:
            config = ret
            save_config(config)
            self.project.analyse_all_models()  # recalc all models
            self.update_data_in_GUI()

    def update_data_in_GUI(self):
        """
        plots all winding data in gui
        """
        bc, bc_str = self.data.get_basic_characteristics()
        self.textBrowser_wdginfo.setHtml(bc_str)
        idx_old = self.comboBox_star_harmonics.currentText()
        idx = (
            self.data.results["nu_el"].index(int(idx_old)) if idx_old else None
        )  # remember ordinal number
        self.comboBox_star_harmonics.blockSignals(
            True
        )  # prevent double plotting on startup
        self.comboBox_star_harmonics.clear()
        self.comboBox_star_harmonics.addItems(
            [str(k) for k in self.data.results["nu_el"]]
        )
        self.comboBox_star_harmonics.blockSignals(False)
        if idx:
            self.comboBox_star_harmonics.setCurrentIndex(
                idx
            )  # restore ordinal number for the new winding
        self.update_plot_in_GUI()
        self.reportEdit.setText(self.data.get_text_report())

    def update_plot_in_GUI(self, small_update=False):
        """
        update the current figure
        if 'small_update' is True, only new lines/bars are plottet and not the 
        all axes etc. --> speed up, for MMK-phase slider for example
        """
        # Update Figures
        if self.plot_tabs.currentWidget():  # if all tabs are closed
            tab_name = self.plot_tabs.currentWidget().objectName()
            if tab_name == "tab_slot":
                self.fig[tab_name].plot_slots(self.data.get_num_slots())
                self.fig[tab_name].plot(
                    self.data, draw_poles=config["plt"]["draw_poles"]
                )
            if tab_name == "tab_layout_polar":
                self.fig[tab_name].plot(
                    self.data,
                    optimize_overhang=self.checkBox_opt_wdg_overhang_layout_polar.isChecked(),
                    draw_poles=config["plt"]["draw_poles"],
                )
            if tab_name == "tab_overhang":
                self.fig[tab_name].plot(
                    self.data,
                    optimize_overhang=self.checkBox_opt_wdg_overhang.isChecked(),
                )
            if tab_name == "tab_star":
                self.fig[tab_name].plot(
                    self.data,
                    harmonic_idx=self.comboBox_star_harmonics.currentIndex(),
                    ForceX=self.checkBoxForceX.isChecked(),
                )
            if tab_name == "tab_wf":
                if self.radioButton_electrical.isChecked():
                    self.fig[tab_name].plot(self.data, mechanical=False)
                elif self.radioButton_mechanical.isChecked():
                    self.fig[tab_name].plot(self.data, mechanical=True)
            if tab_name == "tab_mmk":
                f = _get_float(self.MMK_phase_edit.text())
                f = 0.0 if f is None else f
                self.fig[tab_name].plot(self.data, f, small_update=small_update)

    def printer(self):
        printer = QPrinter(QPrinter.HighResolution)

        dialog = QPrintDialog(printer)

        # get actual view
        tab_name = self.plot_tabs.currentWidget().objectName()
        if tab_name in [
            "tab_slot",
            "tab_layout_polar",
            "tab_overhang",
            "tab_star",
            "tab_wf",
            "tab_mmk",
        ]:
            with tempfile.TemporaryDirectory() as tmpdir:
                self.fig[tab_name].save(
                    os.path.join(tmpdir, "fig.png"), config["plt"]["res"]
                )
                pixmap = QPixmap(os.path.join(tmpdir, "fig.png"))
                if dialog.exec_() == QPrintDialog.Accepted:
                    painter = QPainter(printer)

                    rect = painter.viewport()
                    size = pixmap.size()
                    size.scale(rect.size(), QtCore.Qt.KeepAspectRatio)
                    painter.setViewport(rect.x(), rect.y(), size.width(), size.height())
                    painter.setWindow(pixmap.rect())
                    painter.drawPixmap(0, 0, pixmap)
                    del painter

        elif tab_name == "tab_report":
            if self.reportEdit.textCursor().hasSelection():
                dlg.addEnabledOption(QPrintDialog.PrintSelection)
            if dialog.exec_() == QPrintDialog.Accepted:
                self.reportEdit.print_(printer)
        else:
            print("printing not supported for the current content")
        del dialog

    def plot_tabs_init(self):
        self.plot_tabs_store()
        for k in range(self.plot_tabs.count()):
            self.plot_tabs_remove(0)

        pt = config["view"]["plot_tabs"]
        for p in pt:
            self.plot_tabs_add(p, update_menu=True)
        #  self.plot_tabs.setTabVisible(0, False)
        self.plot_tabs.setCurrentIndex(0)

    def plot_tabs_store(self):
        self.plot_tabs_store = {}
        for k in range(self.plot_tabs.count()):
            tab = self.plot_tabs.widget(k)
            self.plot_tabs_store[tab.objectName()] = {}
            self.plot_tabs_store[tab.objectName()]["widget"] = self.plot_tabs.widget(k)
            self.plot_tabs_store[tab.objectName()]["icon"] = self.plot_tabs.tabIcon(k)
            self.plot_tabs_store[tab.objectName()]["label"] = self.plot_tabs.tabText(k)

    def plot_tabs_update_menu(self, key, flag):
        self.setUpdatesEnabled(False)  # block signals
        self.plot_tabs_actions[key].setChecked(flag)
        self.setUpdatesEnabled(True)  # unblock signals

    def plot_tabs_add(self, key, update_menu=False):
        self.plot_tabs.addTab(
            self.plot_tabs_store[key]["widget"], self.plot_tabs_store[key]["label"]
        )
        idx = self.plot_tabs.count()
        self.plot_tabs.setTabIcon(idx - 1, self.plot_tabs_store[key]["icon"])
        if update_menu:
            self.plot_tabs_update_menu(key, True)
        self.plot_tabs.setCurrentIndex(self.plot_tabs.count() - 1)

    def plot_tabs_remove(self, idx, update_menu=True):
        """
        Remove tabs by index
        """
        key = self.plot_tabs.widget(idx).objectName()
        self.plot_tabs.removeTab(idx)

        #  print('remove', key, idx)
        if update_menu:
            self.plot_tabs_update_menu(key, False)

    def plot_tabs_remove_by_name(self, key, update_menu=True):
        """
        Remove tabs by name
        """
        for k in range(self.plot_tabs.count()):
            widget = self.plot_tabs.widget(k)
            if widget is not None:
                if widget.objectName() == key:
                    self.plot_tabs_remove(k, update_menu=update_menu)

    def plot_tabs_update_from_menu(self):
        """
        remove/add tabs depended on the menu
        """
        checked = []  # checks from menu
        for key, value in self.plot_tabs_actions.items():
            if value.isChecked():
                checked.append(key)

        checked2 = []  # active tabs
        for k in range(self.plot_tabs.count()):
            checked2.append(self.plot_tabs.widget(k).objectName())

        # activate tabs to sync with menu
        for c in checked:
            if c not in checked2:
                self.plot_tabs_add(c)

        # deactivate tabs to sync with menu
        for c in checked2:
            if c not in checked:
                self.plot_tabs_remove_by_name(c)

    def plot_tabs_show_all(self):
        for k in range(self.plot_tabs.count()):
            self.plot_tabs_remove(0)
        default_config = get_config(default=True)
        for tab in default_config["view"]["plot_tabs"]:
            self.plot_tabs_add(tab, update_menu=True)
        self.plot_tabs.setCurrentIndex(0)

    def plot_tabs_save_config(self):
        checked = []
        for k in range(self.plot_tabs.count()):
            checked.append(self.plot_tabs.widget(k).objectName())
        config["view"]["plot_tabs"] = checked
        save_config(config)

    def save_to_file(self):
        """
        returns True if file is saved
        """
        if not self.project.filename:
            ret = self.save_as_to_file()
            return ret
        else:
            self.project.save_to_file(self.project.filename)
            self.statusbar.showMessage(
                "Project saved as " + self.project.filename, MSG_TIME
            )
            self.check_is_saved()
            return True

    def save_as_to_file(self):
        """
        returns True if file is saved
        """
        options = QFileDialog.Options()
        filename, _ = QFileDialog.getSaveFileName(
            self, "Save winding file", "", "Winding Files (*.wdg)", options=options
        )
        if filename:
            if not filename.endswith(".wdg"):
                filename += ".wdg"
            self.project.set_filename(filename)
            self.project.save_to_file(self.project.filename)
            self.statusbar.showMessage(
                "Project saved as " + self.project.filename, MSG_TIME
            )
            self.check_is_saved()
            return True
        else:
            return False

    def load_from_file(self, filename=None):
        """
        loads a *.wdg file in the workspace
        if filename is not given a file dialog appears
        """
        if not filename:
            if not self.project.get_save_state():
                ok = QMessageBox.question(
                    self,
                    "Load *.wdg file",
                    "Do you want to save?",
                    QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel,
                    QMessageBox.Cancel,
                )
                if ok == QMessageBox.Yes:
                    ret = self.save_to_file()
                    if not ret:  # user has not saved
                        return
                if ok == QMessageBox.Cancel:  # cancel
                    return

            options = QFileDialog.Options()
            filename, _ = QFileDialog.getOpenFileName(
                self, "Open winding file", "", "Winding Files (*.wdg)", options=options
            )
        if filename:
            self.project.load_from_file(filename)
            self.statusbar.showMessage("Project loaded: " + filename, MSG_TIME)
            self.project.set_filename(filename)
            self.update_project_list()
            self.update_data_in_GUI()
        self.check_is_saved()

    def tasks_on_exit(self):
        """
        Do some taske when the main window is closed
        """
        self.plot_tabs_save_config()

    def closeEvent(self, event):
        # Exit programm

        # Ask for saving
        if self.project.get_save_state():
            self.tasks_on_exit()
            event.accept()  # let the window close
        else:
            ok = QMessageBox.question(
                self,
                "Exit program",
                "Do you want to save?",
                QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel,
                QMessageBox.Cancel,
            )
            if ok == QMessageBox.Yes:
                ret = self.save_to_file()
                if ret:  # user has saved
                    self.tasks_on_exit()
                    event.accept()  # let the window close
                else:  # user has not saved
                    event.ignore()
            elif ok == QMessageBox.No:
                self.tasks_on_exit()
                event.accept()  # close the window
            else:
                event.ignore()  # let the window open

    def open_manual(self):
        doc = os.path.join(__dir__, "doc", "SWAT-EM_manual.pdf")
        if platform.system() == "Darwin":  # macOS
            subprocess.call(("open", doc))
        elif platform.system() == "Windows":  # Windows
            os.startfile(doc)
        else:  # linux variants
            subprocess.call(("xdg-open", doc))

    def create_report(self):
        rep = report.HtmlReport(self.data)
        rep.create()
        rep.open_in_browser()

    def export_to_excel(self):
        """
        Export to excel
        """
        options = QFileDialog.Options()
        filename, _ = QFileDialog.getSaveFileName(
            self, "Save to Excel file", "", "Excel (*.xlsx)", options=options
        )
        if filename:
            if not filename.endswith(".xlsx"):
                filename += ".xlsx"
            self.data.export_xlsx(filename)

    def export_to_txt(self):
        """
        Export text report
        """
        options = QFileDialog.Options()
        filename, _ = QFileDialog.getSaveFileName(
            self, "Save to Text file", "", "Text (*.txt)", options=options
        )
        if filename:
            if not filename.endswith(".txt"):
                filename += ".txt"
            self.data.export_text_report(filename)

    def set_focus_to_find(self):
        self.plot_tabs.setCurrentWidget(self.tab_report)
        self.lineEdit_find.setFocus()

    def find_in_report(self):
        num_found = self.report.find(self.lineEdit_find.text())
        if num_found == -1:
            self.lineEdit_find.setStyleSheet("background-color: #FF0000")
        else:
            self.lineEdit_find.setStyleSheet("background-color: #FFFFFF")


def main():
    # command line options
    parser = argparse.ArgumentParser(
        prog="SWAT-EM",
        description="Generating and evaluating of windings for electrial machines",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        action="store",
        nargs="?",
        default="",
        dest="arg_1",
        help="*.py scriptfile or *.wdg file",
    )
    parser.add_argument(
        "--load",
        action="store",
        default="",
        dest="loadfile",
        help="Open an existing *.wdg file",
    )
    parser.add_argument(
        "--script",
        action="store",
        default="",
        dest="scriptfile",
        help="Open an existing *.py script file",
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s " + __version__,
        help="Show version",
    )
    args = parser.parse_args()

    if len(args.arg_1) > 0:
        #  print('my arg: ', args.arg_1)
        ext = os.path.splitext(args.arg_1)[-1]
        if ext.upper() == ".PY":
            args.scriptfile = args.arg_1
        elif ext.upper() == ".WDG":
            args.loadfile = args.arg_1

    if args.scriptfile:
        print("eval script file:", args.scriptfile)
    else:
        print("running GUI")
        app = QApplication(sys.argv)
        ex = MainWindow()

        if args.loadfile:
            if os.path.isfile(args.loadfile):
                ex.load_from_file(args.loadfile)
            else:
                raise (Exception, "file not found:".format(args.loadfile))
        splash.close()
        sys.exit(app.exec_())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import swat_em.config

# TODO: Test the config save without overwriting the users config in home folder!
#  def test1():
#  config = swat_em.config.get_config()
#  swat_em.config.save_config(config)
#  config2 = swat_em.config.get_config(init = False)
#  assert config == config2

#  if __name__ == '__main__':
#  test1()



# -*- coding: utf-8 -*-

# Functions for creating windings

import fractions

#  from collections import deque
import math
import numpy as np
from swat_em import analyse


def is_even(val):
    """
    Returns True if absolute value is even
    """
    return True if abs(val) % 2 == 0 else False


def genwdg(Q, P, m, w, layers, empty_slots=0):
    """
    Generates a winding layout.

    Parameters
    ----------
    Q :      integer
             number of slots
    P :      integer
             number of poles
    m :      integer
             number of phases
    w :      integer
             winding step (1 for tooth coils)
    layers : integer
             number of coil sides per slot
    empty_slots : integer
                    Defines the number of empty slots ("dead coil winding")
                    -1: Choose number of empty slots automatically (the smallest
                      possible number is choosen)
                    0: No empty slots
                    >0: Manual defined number of empty slots

             
    Returns
    -------
    return : list
             winding layout (right and left layers) for every phase
    """
    if empty_slots == 0:
        ret = winding_from_star_of_slot(Q, P, m, w, layers)
        if not ret["valid"] and empty_slots != 0:
            ret = winding_from_general_equation(Q, P, m, w, layers, empty_slots)
        if not ret["valid"]:
            ret = None
    elif empty_slots == -1:
        if layers == 1 and Q % (2 * m) == 0:
            ret = winding_from_star_of_slot(Q, P, m, w, layers)
        elif layers == 2 and Q % (m) == 0:
            ret = winding_from_star_of_slot(Q, P, m, w, layers)
        else:
            ret = winding_from_general_equation(Q, P, m, w, layers, empty_slots)
    else:
        ret = winding_from_general_equation(Q, P, m, w, layers, empty_slots)
    return ret


def overlapping_fractional_slot_slayer(Q, P, m):
    from collections import deque

    p = P / 2
    q = fractions.Fraction(Q / (m * 2 * p)).limit_denominator(100)
    g = 0  # number of full slots
    while q > 1:
        q -= 1
        g += 1

    z2 = q.numerator
    n = q.denominator

    # build sequence of coil sides (half of the coil side pairs of the coils)
    cgroups = []
    for k in range(n):
        if k < z2:
            cgroups.append(g + 1)
        if k < (n - z2):
            cgroups.append(g)
    cgroups *= m

    # build sequnce of coil sides for each phase (complete pair)
    cgroups_per_phase = [[] for k in range(m)]
    i = 0
    for g in cgroups:
        cgroups_per_phase[i].append(g)
        cgroups_per_phase[i].append(-g)
        #  print('g', g)
        i += 1
        if i >= m:
            i = 0

    if m == 3:  # change phase sequence for generating the winding layout
        cgroups_per_phase[1], cgroups_per_phase[2] = (
            cgroups_per_phase[2],
            cgroups_per_phase[1],
        )

    for i in range(m):
        if i % 2 != 0:
            l = deque(cgroups_per_phase[i])
            l.rotate(1)
            cgroups_per_phase[i] = list(l)

    # print(cgroups_per_phase)
    # print('z2', z2)
    # print('n', n)
    # print(cgroups)

    phases = [[] for k in range(m)]
    i = 1
    while i < Q:
        for cs in zip(*cgroups_per_phase):
            for km in range(len(cs)):
                for s in range(abs(cs[km])):
                    if cs[km] > 0:
                        phases[abs(km)].append(-i)
                    else:
                        phases[abs(km)].append(i)
                    i += 1
    #  print('phases', phases)
    #  print('cgroups_per_phase', cgroups_per_phase)

    if m == 3:  # change phase sequence again
        phases[1], phases[2] = phases[2], phases[1]

    for k in range(len(phases)):
        phases[k] = [phases[k], []]

    # windingsteps for the fractional slot winding
    w = [0] * len(cgroups_per_phase[0])
    #  for i, cg in enumerate(cgroups_per_phase):
    #  for k in range(len(cg)):
    #  w[k] += abs(cg[k])
    #  print('w[i]', w[i])

    cg2 = list(map(list, zip(*cgroups_per_phase)))  # transpose
    w = []
    for cg in cg2:
        cg = [abs(k) for k in cg]
        w.append(sum(cg))
    #  print('w', w)
    w = w[: len(w) // 2]

    #  print('cg2', cg2)
    #  print('w', w)
    return phases, w


def winding_from_star_of_slot(Q, P, m, w=-1, layers=2):
    """
    Based on:
    N. Bianchi and M. Dai Pre, "Use of the star of slots in designing fractional-slot single-layer synchronous motors," in IEE Proceedings - Electric Power Applications, vol. 153, no. 3, pp. 459-466, 1 May 2006.
    doi: 10.1049/ip-epa:20050284
    keywords: {synchronous motors;machine windings;torque;losses;fault tolerance;harmonics;coils;fractional-slot winding;single-layer synchronous motor;end-winding loss reduction;torque ripple;mutual coupling reduction;fault-tolerant application;one side coil;slot star;graphical representation;analytical formulation;harmonic content},
    URL: http://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=1629527&isnumber=34185
    """
    if hasattr(w, "__iter__"):
        w = -1  # fractional slot has list of w's

    if w == -1:
        w = Q // P
        if w <= 0:
            w = 1

    p = P // 2

    error = ""
    valid = True

    t = math.gcd(Q, p)
    # Test if there are enough slots for the winding
    if layers == 1:
        test1 = fractions.Fraction(Q / (2 * m)).limit_denominator(100)
        if test1.denominator != 1:
            valid = False
            error += "For single layer winding Q/(2*m) must be an integer\n"
    elif layers == 2:
        test1 = fractions.Fraction(Q / m).limit_denominator(100)
        if test1.denominator != 1:
            valid = False
            error += "For double layer winding Q/m must be an integer"

    test2 = fractions.Fraction(Q / (m * t)).limit_denominator(100)
    if test2.denominator != 1:
        valid = False
        error += "winding not feasible"

    phasors = []
    for k in range(Q):
        a = 2 * np.pi * p / Q * k
        a += np.pi / m / 4  # shift by half sector range instead of shifting
        a -= 2 * np.pi * p / Q / 100  # add a small shift because a phasor could match
        # a sector border exactly

        while a > 2 * np.pi:  # angle between 0...2pi
            a -= 2 * np.pi
        phasors.append(a)

    phases = [[[], []] for k in range(m)]
    for km in range(m):  # for every phase
        r = np.pi / m  # range for sector
        if is_even(m):  # shifting m=2 (and maybe m=4, 6,...)
            mp = 1  # with special treatment
        else:
            mp = 2
        r1 = [mp * km * r, mp * km * r + r]  # sector for positive coils sides
        r2 = [r1[0] + np.pi, r1[1] + np.pi]  # sector for negative coils sides

        for k in range(len(r1)):  # sector between 0..2pi
            while r1[k] > 2 * np.pi:
                r1[k] -= 2 * np.pi
        for k in range(len(r2)):
            while r2[k] > 2 * np.pi:
                r2[k] -= 2 * np.pi

        for i, p in enumerate(phasors):
            # print(i+1, p, r1, r2)
            if p > r1[0] and p <= r1[1]:  # phasor is in positive sector
                phases[km][0].append(i + 1)  # slot belongs to the phase
                c = i + 1 + w
                while c > Q:
                    c -= Q
                phases[km][1].append(-c)  # add the second layer
            if p > r2[0] and p <= r2[1]:  # phasor is in negative sector
                phases[km][0].append(-i - 1)
                c = i + 1 + w
                while c > Q:
                    c -= Q
                phases[km][1].append(c)

    # remove second layer. Attention: It is important to fill the second
    # layer with the algorithm above, even if single layer is to generate,
    # because there are special cases, where the single layer winding have
    # to be generated from the double layer winding!
    if layers == 1:  # only
        if not is_even(w):
            for km in range(m):
                a = phases[km]
                a2 = [[], []]
                for k in range(len(a[0])):
                    if not is_even(a[0][k]):  # use only the odd slots
                        a2[0].append(a[0][k])
                        a2[0].append(a[1][k])  # from second layer to first layer
                phases[km] = a2
        else:
            # use fallback function for w = even!
            q = fractions.Fraction(Q / (m * P)).limit_denominator(100)
            if q.denominator == 1:
                for km in range(m):
                    phases[km][1] = []
            else:
                #  print('single layer winding not suitable', Q, P)
                phases, w = overlapping_fractional_slot_slayer(Q, P, m)
                #  w = Q/P

    if Q % P != 0 and w != 1:
        w = [Q // P, Q // P + 1]

    ret = {
        "phases": phases,
        "wstep": w,
        "valid": valid,
        "error": error,
        "info": "",
        "Qes": 0,
    }
    return ret


def winding_from_general_equation(Q, P, m, w=-1, layers=2, n_es=0):
    """
    Based on:
    Caruso, Massimo & Di Tommaso, Antonino & Marignetti, Fabrizio & Miceli, Rosario & Galluzzo, Giuseppe. (2018). A General Mathematical Formulation for Winding Layout Arrangement of Electrical Machines. Energies. 11. 446. 10.3390/en11020446. 
    """
    error = ""
    info = ""
    valid = True
    S = [[[]] * m]
    if layers == 1 and w == 1:
        info += "Coil pitch w = 1 could not be applied"

    N = Q
    p = P // 2
    n_lay = layers
    if layers == 2 and w == -1:
        w = Q // P
        if w <= 0:
            w = 1

    if m != 1 and Q % m != 0:
        valid = False
        error += "Number of slots must be a multiple of the number of phases"

    if valid:
        if n_es >= 0:
            n_wc = n_lay * (N - n_es) / (2 * m)
        else:
            n_wc = n_lay * (N - 0) / (2 * m)
        t = math.gcd(N, p)

        # symmetric winding?
        if m % 2 == 0:
            g = n_lay * N / (2 * m * t)
        else:
            g = N / (2 * m * t)
        #  if int(g) != g:
        #  if int(n_wc) != n_wc:
        #  valid = False
        #  error += 'winding not symmetric'

    if valid:
        # dead coil winding (empty slots)
        if n_es == -1:
            n_es = int(N - 2 * m * int(n_wc) / n_lay)

        if n_es != 0:
            info += "attention: dead coil winding"
        #  print('empty_slots:', n_es)
        q = fractions.Fraction((N - n_es) / (2 * p * m)).limit_denominator(1000)
        a = int(q)
        z = q.numerator - a

    if valid:
        # create winding distribution table
        n_c = int(N / m)
        #  print('Q', Q, 'p', p, 'n_c', n_c)
        WDT = np.zeros(m * n_c, dtype=int)
        i = 1
        while i <= N:
            ind = np.mod(p * (i - 1) + 1, N)
            if ind == 0:
                ind = N
            while WDT[ind - 1] != 0:
                ind = ind + 1
            WDT[ind - 1] = i
            i = i + 1
            ind = ind + 1

        WDT = WDT.reshape(m, n_c)

        if m % 2 == 0:
            shift = int(m / 2 - 1)
        else:
            shift = int((m - 1) / 2)

    # for some bar windings
    #  if fractions.Fraction(n_wc).limit_denominator(1000).denominator == 2:
    #  shift = n_wc + 1 # this
    #  shift = n_wc - 1 # or this is possible

    if valid:
        #  print('Q', Q, 'p', p, 'n_es', n_es, 'n_c', n_c)
        # non-ruduced and normal systems
        if m % 2 != 0:
            a = WDT[:, : int(n_c / 2)]
            b = WDT[:, int(n_c / 2) :] * (-1)
            b = np.roll(b, -shift, axis=0)
            WDT2 = np.append(a, b, axis=1)
        else:
            dx = int(n_c / 2)
            dy = int(m / 2)
            a = WDT[:dy, :dx]
            b = WDT[:dy, dx:] * (-1)
            c = WDT[dy:, :dx] * (-1)
            d = WDT[dy:, dx:]

            try:
                left = np.append(a, b, axis=0)
                right = np.append(c, d, axis=0)
                WDT2 = np.append(left, right, axis=1)
            except:
                # Error
                WDT2 = WDT
                valid = False
                error += "winding not feasable"

        if n_es > 0:
            WDT2 = WDT2[:, : -int(n_es / m)]

        S = []
        for k in range(np.shape(WDT2)[0]):
            s = WDT2[k, :]
            idx = np.argsort(np.abs(s))
            s = s[idx]
            S.append([s.tolist(), []])

        if n_lay == 2:
            for k in range(len(S)):
                for s in S[k][0]:
                    sign = 1 if s > 0 else -1
                    s = abs(s) + w
                    while s > N:
                        s -= N
                    while s < 1:
                        s += N
                    S[k][1].append(sign * (-1) * s)

        if n_lay == 1:
            w = fractions.Fraction(N / (2 * p))

    if valid:
        v, x = analyse.check_number_of_coilsides(S)
        if not v:
            valid = False
        error += x

    if Q % P != 0 and w != 1:
        w = [Q // P, Q // P + 1]

    ret = {
        "phases": S,
        "wstep": w,
        "valid": valid,
        "error": error,
        "info": info,
        "Qes": n_es,
    }
    return ret



# -*- coding: utf-8 -*-

import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import numpy as np
from swat_em import datamodel


wdg = datamodel()

#  wdg.genwdg(Q=12, P=10, m=3, w=1, layers=2) # passt
#  wdg.genwdg(Q=12, P=10, m=3, w=1, layers=1) # passt


#  wdg.genwdg(Q=6, P=2, m=3, w=-1, layers=1)
#  wdg.genwdg(Q=12, P=2, m=3, w=-1, layers=1)
#  wdg.genwdg(Q=12, P=2, m=3, w=6, layers=2)
#  wdg.genwdg(Q=12, P=2, m=3, w=5, layers=2)

#  wdg.genwdg(Q=12, P=4, m=3, w=-1, layers=1)

#  wdg.genwdg(Q=9, P=2, m=3, w=-1, layers=2)

wdg.genwdg(Q=18, P=4, m=3, w=-1, layers=1)  # fractional slot

#  plot_overhang(self, filename, res = None, show = False, optimize_overhang = False)
wdg.plot_overhang("overhang.png")


#  S = [[1, 2, -7, -8], [5, 6, -11, -12], [-3, -4, 9, 10]]
#  wdg.set_machinedata(Q=12, p=1, m=3)
#  wdg.set_phases(S)



# -*- coding: utf-8 -*-
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
import pdb

from swat_em import datamodel


def test1():
    print("Test saving and loading of data files inc. results and options")
    data = datamodel()
    data.set_machinedata(Q=12, p=5, m=3)

    U = [[1, 8], [-2, -7]]
    V = [[4, 9], [-3, -10]]
    W = [[5, 12], [-6, -11]]

    data.set_phases([U, V, W])
    data.analyse_wdg()

    data.save_to_file("savefile.wdg")
    data2 = datamodel()
    data2.load_from_file("savefile.wdg")
    os.remove("savefile.wdg")

    assert data.machinedata == data2.machinedata
    #  assert data.results == data2.results


if __name__ == "__main__":
    test1()



# -*- coding: utf-8 -*-

from PyQt5 import uic
from PyQt5.QtWidgets import QDialog, QMessageBox
import sys
import os
from swat_em import config

if getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS"):
    __dir__ = sys._MEIPASS  # for pyinstaller
else:
    __dir__ = os.path.dirname(os.path.abspath(__file__))


class Settings(QDialog):
    def __init__(self, config):
        self.config = config
        super().__init__()
        #  self.resetted = False # flag is set, if user reset the config

        # Set up the user interface from Designer.
        uic.loadUi(os.path.join(__dir__, "ui", "Settings.ui"), self)
        self.setWindowTitle("Settings")

        #  self.Button_CANCEL.clicked.connect(self.reject)
        #  self.Button_OK.clicked.connect(self.accept)
        self.Button_RESET.clicked.connect(self.reset)

        # Nothing is done, when a value is changed, yet
        # later here can be checked for plausibility

        #  self.spinBox_nu_el.valueChanged.connect(self.update)
        #  self.spinBox_nu_mech.valueChanged.connect(self.update)
        #  self.spinBox_nu_MMK.valueChanged.connect(self.update)
        #  self.doubleSpinBox_threshold_wf.valueChanged.connect(self.update)

    def reset(self):
        ok = QMessageBox.question(
            self,
            "Reset settings",
            "Do you want to reset the settings to default values?",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )
        if ok == QMessageBox.Yes:
            #  self.resetted = True
            self.config = config.get_init_config()
            self.write_to_gui()

    # Nothing is done, when a value is changed, yet
    # later here can be checked for plausibility
    #  def update(self):
    #  print('update')

    def write_to_gui(self):
        """
        Write the config values to the cui
        """
        self.spinBox_nu_el.setValue(self.config["N_nu_el"])
        self.spinBox_nu_mech.setValue(self.config["N_nu_mech"])
        self.doubleSpinBox_threshold_MMK.setValue(
            self.config["threshold_MMF_harmonics"]
        )
        self.doubleSpinBox_threshold_wf.setValue(self.config["kw_min"])
        self.doubleSpinBox_threshold_MMF_plot.setValue(
            self.config["plot_MMF_harmonics"]
        )
        self.spinBox_num_modes_radial_force.setValue(
            self.config["radial_force"]["num_modes"]
        )

        #  self.spinBox_DPI.setValue(self.config['plt']['DPI'])
        self.spinBox_resx.setValue(self.config["plt"]["res"][0])
        self.spinBox_resy.setValue(self.config["plt"]["res"][1])

        self.doubleSpinBox_lw.setValue(self.config["plt"]["lw"])
        self.doubleSpinBox_lw_thin.setValue(self.config["plt"]["lw_thin"])
        self.checkBox_draw_poles.setChecked(self.config["plt"]["draw_poles"])

    def read_from_gui(self):
        """
        Read the config values from the cui
        """
        self.config["N_nu_el"] = self.spinBox_nu_el.value()
        self.config["N_nu_mech"] = self.spinBox_nu_mech.value()
        self.config[
            "threshold_MMF_harmonics"
        ] = self.doubleSpinBox_threshold_MMK.value()
        self.config["kw_min"] = self.doubleSpinBox_threshold_wf.value()
        self.config[
            "plot_MMF_harmonics"
        ] = self.doubleSpinBox_threshold_MMF_plot.value()
        self.config["radial_force"][
            "num_modes"
        ] = self.spinBox_num_modes_radial_force.value()

        #  self.config['plt']['DPI'] = self.spinBox_DPI.value()
        self.config["plt"]["res"][0] = self.spinBox_resx.value()
        self.config["plt"]["res"][1] = self.spinBox_resy.value()

        self.config["plt"]["lw"] = self.doubleSpinBox_lw.value()
        self.config["plt"]["lw_thin"] = self.doubleSpinBox_lw_thin.value()
        self.config["plt"]["draw_poles"] = self.checkBox_draw_poles.isChecked()

    def run(self):
        self.write_to_gui()

        ok = self.exec_()
        if ok:
            self.read_from_gui()
            return self.config



from PyQt5 import uic
from PyQt5.QtWidgets import QDialog
import os
import sys

if getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS"):
    __dir__ = sys._MEIPASS  # for pyinstaller
else:
    __dir__ = os.path.dirname(os.path.abspath(__file__))
import swat_em


class about(QDialog):
    def __init__(self):
        super().__init__()
        # Set up the user interface from Designer.
        uic.loadUi(os.path.join(__dir__, "ui", "About.ui"), self)
        self.Button_close.clicked.connect(self.close)

        txt = self.textBrowser.toHtml()
        txt = txt.replace("%1%", swat_em.__version__)  # placeholder in Qt Designer
        self.textBrowser.setHtml(txt)

        self.exec_()



# -*- coding: utf-8 -*-
"""
Provides functions for reports
"""
import os
import tempfile
import numpy as np
import xlsxwriter
import re
from swat_em import analyse
from swat_em.config import config


def italic(txt):
    return "<i>" + txt + "</i>"


def table_header(data):
    txt = ["<tr>"]
    for d in data:
        txt.append("<th>")
        txt.append(d)
        txt.append("</th>")
    txt.append("</tr>")
    return "".join(txt)


def table_row(data, col=None):
    if col:
        txt = ['<tr bgcolor="{}">'.format(col)]
    else:
        txt = ["<tr>"]
    for d in data:
        txt.append("<td>")
        txt.append(d)
        txt.append("</td>")
    txt.append("</tr>")
    return "".join(txt)


def red(text):
    txt = ['<span style=" color:#ff0000;" >']
    txt.append(text)
    txt.append("</span>")
    return "".join(txt)


def italic(text):
    return "<i>" + text + "</i>"


def bold(text):
    return "<b>" + text + "</b>"


def table(dat, header=[]):
    txt = ['<table style="width:100%">']
    if len(header) > 0:
        txt.append(table_header(header))
    col1 = "#FFFFFF"
    col2 = "#DDDADA"
    for i, d in enumerate(dat):
        col = col1 if i % 2 != 0 else col2
        txt.append(table_row(d, col=col))

    txt.append("</table>")
    return txt


def num2str(number, maxlen=8):
    """
    Converts a number to string with max number of signs

    Parameters
    ----------
    number : scalar, string
             Number to convert
    maxlen : integer
             Max. length to convert
           
    Returns
    -------
    return : string / list of strings / ndarray of strings
             string with the number           
    """
    if number > 0:
        maxlen -= 1
    numstring = str(number)
    ematch = re.search("[eE].+", numstring)
    if ematch:
        # number in scientific
        exp = ematch.group()
        length = min(len(numstring), maxlen) - len(exp)
        numstring = numstring[:length] + exp
    else:
        numstring = numstring[:maxlen]
    if number >= 0.0:
        numstring = " " + numstring
    return numstring


class HtmlReport:
    def __init__(self, data):
        """
        Class for creating HTML reports of the winding

        Parameters
        ----------
        data :   datamodel object
                 winding model for creating the report
        """
        self.data = data
        self.template = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "template", "report.html")
        )

    def _create_figures(self, tmpdir):
        """
        Creates all figures for the winding
        
        Parameters
        ----------
        tmpdir : string
                 name of the directory to store the figures
        """
        if not os.path.isdir(tmpdir):
            os.mkdir(tmpdir)
        self.data.plot_layout(
            os.path.join(tmpdir, "plot_layout.png"), res=config["plt"]["res"]
        )
        self.data.plot_star(
            os.path.join(tmpdir, "plot_star.png"), res=config["plt"]["res"]
        )
        self.data.plot_windingfactor(
            os.path.join(tmpdir, "plot_wf_el.png"),
            mechanical=False,
            res=config["plt"]["res"],
        )
        self.data.plot_windingfactor(
            os.path.join(tmpdir, "plot_wf_mech.png"),
            mechanical=True,
            res=config["plt"]["res"],
        )
        self.data.plot_MMK(
            os.path.join(tmpdir, "plot_MMK.png"), res=config["plt"]["res"], phase=0
        )

    def create(self, filename=None):
        """
        Generates the html-report and returns the name of the html file

        Returns
        -------
        return : string
                 The file name of the html-file which is stored in tmp directory
        """
        if filename == None:
            self.tmpdir = tempfile.mkdtemp()
            self.filename = os.path.join(self.tmpdir, "report.html")
        else:
            self.tmpdir = os.path.abspath(os.path.dirname(filename))
            self.filename = filename
        self.contentdir = os.path.splitext(self.filename)[0] + "-files"
        try:
            os.mkdir(self.contentdir)
        except:
            pass

        #  filename = '/media/ramdisk/report.html'
        #  tmpdir = '/media/ramdisk'   # for testing

        self._create_figures(self.contentdir)

        # create table of star of slots
        d = []
        ei = self.data.results["Ei_el"][0]
        for km in range(len(ei)):
            A = np.abs(np.sum(ei[km]))
            ph = np.angle(np.sum(ei[km])) / np.pi * 180
            d.append([str(km + 1), str(round(A, 1)), str(round(ph, 1))])
        _table_star = "\n".join(table(d, header=["m", "A", "phase"]))

        # create table of winding factor
        # electrical
        d = []
        nu = np.array(self.data.results["nu_el"])
        kw = np.array(self.data.results["kw_el"])
        header = ["nu"] + self.data.get_phasenames()
        for i, line in enumerate(kw):
            d.append([str(nu[i])] + [str(round(k, 3)) for k in line])
        _table_wf_el = "\n".join(table(d, header=header))

        # mechanical
        d = []
        nu = np.array(self.data.results["nu_mech"])
        kw = np.array(self.data.results["kw_mech"])
        header = ["nu"] + self.data.get_phasenames()
        for i, line in enumerate(kw):
            d.append([str(nu[i])] + [str(round(k, 3)) for k in line])
        _table_wf_mech = "\n".join(table(d, header=header))

        # create MMK
        d = []
        phi, MMK, theta = analyse.calc_MMK(
            self.data.get_num_slots(),
            self.data.get_num_phases(),
            self.data.get_phases(),
            self.data.get_turns(),
            angle=0.0,
        )
        nu = np.array(self.data.results["MMK"]["nu"])
        HA = analyse.DFT(MMK[:-1])[: len(nu)]
        A = np.abs(HA)
        header = ["nu", "Amp", "[%]"]
        for k in range(len(nu)):
            a = A[k]
            a_rel = 100.0 / max(A) * a
            d.append([str(nu[k]), str(round(a, 3)), str(round(a_rel, 1))])
        _table_MMK = "\n".join(table(d, header=header))

        # Write to template
        with open(self.template) as f:
            tmpl = f.readlines()

        _report = []
        for line in tmpl:
            line = line.strip()
            line = line.replace("{{ TITLE }}", self.data.get_title())
            line = line.replace("{{ NOTES }}", self.data.get_notes())

            line = line.replace(
                "{{ plot_layout }}", os.path.join(self.contentdir, "plot_layout.png")
            )
            txt, html = self.data.get_basic_characteristics()
            line = line.replace("{{ table_bc }}", html)

            line = line.replace(
                "{{ plot_star }}", os.path.join(self.contentdir, "plot_star.png")
            )
            line = line.replace("{{ table_star }}", _table_star)
            line = line.replace(
                "{{ plot_wf_el }}", os.path.join(self.contentdir, "plot_wf_el.png")
            )
            line = line.replace("{{ table_wf_el }}", _table_wf_el)
            line = line.replace(
                "{{ plot_wf_mech }}", os.path.join(self.contentdir, "plot_wf_mech.png")
            )
            line = line.replace("{{ table_wf_mech }}", _table_wf_mech)
            line = line.replace(
                "{{ plot_MMK }}", os.path.join(self.contentdir, "plot_MMK.png")
            )
            line = line.replace("{{ table_MMK }}", _table_wf_mech)

            _report.append(line)

        with open(self.filename, "w", encoding="utf-8") as f:
            f.write("\n".join(_report))
        return self.filename

    def open_in_browser(self):
        """
        Open the html report in the webbrowser
        """
        if not os.path.isfile(self.filename):
            raise Exception('No html report available. Use the ".create()" before')
        else:
            import webbrowser

            webbrowser.open(self.filename)


def export_xlsx(fname, data):
    layout = True
    phasors = True
    windingfactor_el = True
    windingfactor_mech = True

    workbook = xlsxwriter.Workbook(fname)

    if layout:
        worksheet = workbook.add_worksheet("layout")

        i = 0
        cell_format = workbook.add_format({"font_size": 14})
        worksheet.write(i, 0, "Windinglayout", cell_format)
        i += 1

        cell_format = workbook.add_format({"align": "center"})
        for k in range(data.machinedata["Q"]):
            worksheet.write(i, 1 + k, k + 1, cell_format)
        i += 1

        cell_format = workbook.add_format({"align": "right"})
        worksheet.write(i - 1, 0, "slot No.", cell_format)
        for k in range(data.get_num_layers()):
            worksheet.write(i + k, 0, "Layer " + str(k + 1))

        l, ls, lcol = data.get_layers()
        for k1 in range(np.shape(l)[0]):
            for k2 in range(np.shape(l)[1]):
                cell_format = workbook.add_format(
                    {"bg_color": lcol[k1, k2], "align": "center"}
                )
                worksheet.write(k1 + 2, k2 + 1, ls[k1, k2], cell_format)

        worksheet.set_column(0, 0, width=15)
        worksheet.set_column(1, data.machinedata["Q"], width=4.5)
        i += data.get_num_layers()
        i += 2

        if type(data.machinedata["turns"]) == type([]):
            cell_format = workbook.add_format({"font_size": 14})
            worksheet.write(i, 0, "Number of turns", cell_format)
            i += 1
            for k in range(data.get_num_layers()):
                worksheet.write(i + k, 0, "Layer " + str(k + 1))

            turns = data.machinedata["turns"]
            for km, ph in enumerate(data.machinedata["phases"]):
                col = get_phase_color(km)
                cell_format = workbook.add_format({"bg_color": col, "align": "center"})
                for kl in range(len(ph)):
                    layer = ph[kl]
                    for j, cs in enumerate(layer):
                        if type(turns) == type([]):
                            item = turns[km][kl][j]
                        else:
                            item = turns
                        if cs > 0:
                            worksheet.write(i + kl, 1 + cs - 1, item, cell_format)
                        else:
                            worksheet.write(i + kl, 1 + abs(cs) - 1, item, cell_format)

            i += data.get_num_layers()
        else:
            worksheet.write(i, 0, "Number of turns")
            worksheet.write(i, 1, data.machinedata["turns"])
        i += 2

        cell_format = workbook.add_format({"font_size": 14})
        worksheet.write(i, 0, "slot number per phase", cell_format)
        i += 1
        for kph, ph in enumerate(data.machinedata["phases"]):
            for klay, lay in enumerate(ph):
                worksheet.write(
                    i, 0, "Phase " + str(kph + 1) + " layer " + str(klay + 1)
                )
                cell_format = workbook.add_format({"align": "center"})
                worksheet.write_row(i, 1, lay, cell_format)
                i += 1

    if phasors:
        worksheet = workbook.add_worksheet("phasors")

        cell_format = workbook.add_format({"font_size": 14})
        worksheet.write(0, 0, "Voltage phasors", cell_format)
        max_i = 0
        for knu in range(len(data.results["Ei_el"])):
            i = max_i + 2
            worksheet.write(i, 0, "nu = " + str(data.results["nu_el"][knu]))
            i += 1
            dx = 0

            for m, phasors in enumerate(data.results["Ei_el"][knu]):
                worksheet.write(i, dx + 0, "Phase" + str(m + 1) + "_real")
                worksheet.write(i, dx + 1, "Phase" + str(m + 1) + "_imag")
                for k in range(len(phasors)):
                    worksheet.write(i + 1, dx, phasors[k].real)
                    worksheet.write(i + 1, dx + 1, phasors[k].imag)
                    i += 1
                max_i = i if i > max_i else max_i
                i -= len(phasors)  # go back for new columns
                dx += 2

    if windingfactor_el:
        worksheet = workbook.add_worksheet("Winding_factor_el)")
        cell_format = workbook.add_format({"font_size": 14})
        worksheet.write(0, 0, "Winding factor (electrical)", cell_format)

        worksheet.write(2, 0, "nu")
        for km in range(data.get_num_phases()):
            worksheet.write(2, km + 1, "phase" + str(km + 1))

        nu, kw = data.get_windingfactor_el()
        worksheet.write_column(3, 0, nu)
        for km in range(data.get_num_phases()):
            worksheet.write_column(3, km + 1, kw[:, km])

    if windingfactor_mech:
        worksheet = workbook.add_worksheet("Winding_factor_mech)")
        cell_format = workbook.add_format({"font_size": 14})
        worksheet.write(0, 0, "Winding factor (mechanical)", cell_format)

        worksheet.write(2, 0, "nu")
        for km in range(data.get_num_phases()):
            worksheet.write(2, km + 1, "phase" + str(km + 1))

        nu, kw = data.get_windingfactor_mech()
        worksheet.write_column(3, 0, nu)
        for km in range(data.get_num_phases()):
            worksheet.write_column(3, km + 1, kw[:, km])

    workbook.close()


class TextReport:
    def __init__(self, data):
        """
        Class for creating Text reports of the winding

        Parameters
        ----------
        data :   datamodel object
                 winding model for creating the report
        """
        self.data = data
        self._txt = []

    def create(self):
        self._txt.append("TITLE")
        self._txt.append("=====")
        self._txt.append(self.data.get_title())
        self._txt.append("\n")

        self._txt.append("NOTES")
        self._txt.append("=====")
        self._txt.append(self.data.get_notes())
        self._txt.append("\n")

        self._txt.append("MACHINE DATA")
        self._txt.append("============")
        self._txt.append(
            "Number of slots        Q: {}".format(self.data.get_num_slots())
        )
        self._txt.append(
            "Number of pole pairs   p: {}".format(self.data.get_num_polepairs())
        )
        self._txt.append(
            "Number of phases       m: {}".format(self.data.get_num_phases())
        )
        self._txt.append("Number of slots/phase  q: {}".format(self.data.get_q()))
        self._txt.append("\n")

        _, ls, _ = self.data.get_layers()
        self._txt.append("WINDING LAYOUT")
        self._txt.append("==============")
        self._txt.append("Layer_1: " + " | ".join([k for k in ls[0, :]]))
        if self.data.get_num_layers() > 1:
            self._txt.append("Layer_2: " + " | ".join([k for k in ls[1, :]]))

        self._txt.append("\n")

        bc, txt = self.data.get_basic_characteristics()
        self._txt.append("WINDING FACTOR")
        self._txt.append("==============")
        self._txt.append("Periodic base winding                  t: {}".format(bc["t"]))
        a_ = [str(i) for i in analyse.Divisors(bc["a"])]
        self._txt.append(
            "Possible parallel winding connections  a: {}".format(", ".join(a_))
        )

        for k in range(len(bc["kw1"])):
            self._txt.append(
                "Fundamental winding factor           kw1: {} (phase{})".format(
                    round(bc["kw1"][k], 3), k + 1
                )
            )
        self._txt.append(
            "Double linked leakage            sigma_d: {} (from MMF)".format(
                round(bc["sigma_d"], 3)
            )
        )
        self._txt.append("\n")

        self._txt.append("Harmonic content of the winding faktor (electrical)")
        self._txt.append(
            "nu_el\t"
            + "\t".join(["kw (" + k + ")" for k in self.data.get_phasenames()])
        )
        nu, kw = self.data.get_windingfactor_el()

        for knu in range(len(nu)):
            kw_ = [num2str(k, 8) for k in kw[knu]]
            self._txt.append(str(nu[knu]) + "\t" + "\t".join(kw_))
        self._txt.append("\n")

        self._txt.append("Harmonic content of the winding faktor (mechanical)")
        self._txt.append(
            "nu_mech\t"
            + "\t".join(["kw (" + k + ")" for k in self.data.get_phasenames()])
        )
        nu, kw = self.data.get_windingfactor_mech()

        for knu in range(len(nu)):
            kw_ = [num2str(k, 8) for k in kw[knu]]
            self._txt.append(str(nu[knu]) + "\t" + "\t".join(kw_))
        self._txt.append("\n")

        self._txt.append("MAGNETO MOTIVE FORCE / MMF")
        self._txt.append("==========================")

        threshold = config["threshold_MMF_harmonics"]
        nu, Cnu, phase = self.data.get_MMF_harmonics(threshold=threshold)
        Cnu_rel = 100 / np.max(Cnu) * Cnu

        self._txt.append(
            "Harmonic content of the MMF (airgap field caused by the winding)"
        )
        self._txt.append(
            "Harmonics lower than {}% of the max. amplitude are not shown".format(
                threshold * 100
            )
        )
        self._txt.append("nu_mech\tA^\tArel in %\tphase in °")

        for knu in range(len(nu)):
            self._txt.append(
                str(nu[knu])
                + "\t"
                + num2str(Cnu[knu])
                + "\t"
                + str(round(Cnu_rel[knu], 1))
                + "\t"
                + str(round(phase[knu] / np.pi * 180, 1))
            )

        self._txt.append("\n")

        self._txt.append("FORCES")
        self._txt.append("======")
        self._txt.append(
            "Cogging torque ordinal numbers for pm-machines: {}, {}, {}, ...".format(
                bc["lcmQP"], bc["lcmQP"] * 2, bc["lcmQP"] * 3
            )
        )
        r_ = [str(i) for i in bc["r"]]
        self._txt.append(
            "The winding leads to radial forces with modes r: {}, ...".format(
                ", ".join(r_)
            )
        )

        #  self._txt.append()

        #  self._txt.append()
        #  self._txt.append()

    def get_report(self):
        if len(self._txt) == 0:
            self.create()
        return "\n".join(self._txt)

    def save(self, fname):
        with open(fname, "w") as f:
            f.write(self.get_report())



# -*- coding: utf-8 -*-
# Test detection of the max. possible parallel connection of a winding
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from swat_em.wdggenerator import genwdg
from swat_em.datamodel import datamodel
import swat_em.analyse
import numpy as np
import pdb


def test_parallel_circuit():
    Q = [12, 12, 9, 12, 27]
    P = [10, 8, 8, 2, 18]
    m = [3, 3, 3, 3, 3]
    wstep = [1, 1, 1, 6, 1]
    layers = [2, 2, 2, 1, 2]
    a_desired = [2, 4, 1, 2, 9]

    print("test for max. number of parallel circuits")
    for k in range(len(Q)):
        print(
            "test with Q = {}, 2p = {}, m = {}, wstep = {}, layers = {}".format(
                Q[k], P[k], m[k], wstep[k], layers[k]
            )
        )
        ret = genwdg(Q[k], P[k], m[k], wstep[k], layers[k])

        data = datamodel()
        data.set_machinedata(Q=Q[k], p=P[k] / 2, m=m[k])
        data.set_phases(ret["phases"])
        bc, txt = data.get_basic_characteristics()
        assert a_desired[k] == bc["a"]


if __name__ == "__main__":
    test_parallel_circuit()



# -*- coding: utf-8 -*-
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
from swat_em.datamodel import datamodel


def test_dead_coil_winding():
    print("test dead coil winding")
    wdg = datamodel()
    wdg.genwdg(Q=27, P=6, m=3, layers=1, empty_slots=-1)
    #  print('valid:', wdg.generator_info['valid'])
    #  print('error:', wdg.generator_info['error'])
    #  print('info:', wdg.generator_info['info'])

    S = wdg.get_phases()
    U = [1, 2, -6, 10, -14, -15, 19, -23]
    V = [4, 5, -9, 13, -17, -18, 22, -26]
    W = [-3, 7, 8, -11, -12, 16, -20, 25]

    assert S[0][0] == U
    assert S[1][0] == V
    assert S[2][0] == W


if __name__ == "__main__":
    test_dead_coil_winding()



# -*- coding: utf-8 -*-

import os

__title__ = "SWAT-EM - Specific Winding Analyse Tool for electrical machines"
__version__ = "0.6.0"
__author__ = "Martin Baun"
__license__ = "MIT License"
__copyright__ = "Copyright 2018-2019 Martin Baun"

from swat_em.datamodel import datamodel, project



# -*- coding: utf-8 -*-

import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

#  from swat.wdggenerator import genwdg, overlapping_fractional_slot_slayer, overlapping_fractional_slot_dlayer
from swat_em.wdggenerator import genwdg
from swat_em.datamodel import datamodel
import swat_em.analyse
import numpy as np
import pdb


def test1():
    print("Test if MMK results in the same windingfactor as star of voltage")
    print("double layer winding, tooth-coil")
    Q = 18
    p = 8
    m = 3
    wstep = 1
    layers = 2
    ret = genwdg(Q, 2 * p, m, wstep, layers)
    wdglayout = ret["phases"]
    data = datamodel()
    data.set_machinedata(Q=Q, p=p, m=m)
    data.set_phases(wdglayout)
    data.analyse_wdg()

    nu = data.results["MMK"]["nu"]
    HA = data.results["MMK"]["HA"]
    idx = nu.index(p)

    Ns = Q * layers / m / 2  # number of turns in series
    kw1 = np.abs(HA[idx]) * np.pi * p / (3.0 * Ns)  # winding-factor based on the MMK

    idx = data.results["nu_el"].index(1)
    assert np.round(kw1, 4) == np.round(data.results["kw_el"][idx][0], 4)  # phase U


def test2():
    print("Test if MMK results in the same windingfactor as star of voltage")
    print("single layer winding, tooth-coil")
    Q = 12
    p = 5
    m = 3
    wstep = 1
    layers = 1
    ret = genwdg(Q, 2 * p, m, wstep, layers)
    wdglayout = ret["phases"]
    data = datamodel()
    data.set_machinedata(Q=Q, p=p, m=m)
    data.set_phases(wdglayout)
    data.analyse_wdg()

    nu = data.results["MMK"]["nu"]
    HA = data.results["MMK"]["HA"]
    idx = nu.index(p)

    Ns = Q * layers / m / 2  # number of turns in series
    kw1 = np.abs(HA[idx]) * np.pi * p / (3.0 * Ns)  # winding-factor based on the MMK

    idx = data.results["nu_el"].index(1)
    assert np.round(kw1, 4) == np.round(data.results["kw_el"][idx][0], 4)  # phase U


def test3():
    print("Test if MMK results in the same windingfactor as star of voltage")
    print("double layer winding, overlapping")
    Q = 18
    p = 1
    m = 3
    wstep = 6
    layers = 2
    ret = genwdg(Q, 2 * p, m, wstep, layers)
    wdglayout = ret["phases"]
    data = datamodel()
    data.set_machinedata(Q=Q, p=p, m=m)
    data.set_phases(wdglayout)
    data.analyse_wdg()

    nu = data.results["MMK"]["nu"]
    HA = data.results["MMK"]["HA"]
    idx = nu.index(p)

    Ns = Q * layers / m / 2  # number of turns in series
    kw1 = np.abs(HA[idx]) * np.pi * p / (3.0 * Ns)  # winding-factor based on the MMK

    idx = data.results["nu_el"].index(1)
    assert np.round(kw1, 4) == np.round(data.results["kw_el"][idx][0], 4)  # phase U


if __name__ == "__main__":
    test1()
    test2()
    test3()
